import os
import re
from nicegui import ui
import json


last_mod_time = None
current_json_objects = []
active_ix = None
active_item = None

def set_item(ix):
    global active_item, active_ix
    if active_ix != ix:
        active_ix = ix
        active_item = current_json_objects[ix]
        my_stuff.refresh()
        details.refresh()

def format_time(t):
    if t > 1:
        return f"{t:.1f}s"
    elif t > 0.001:
        return f"{t*1000:.1f}ms"
    elif t > 0.0001:
        return f"{t*1000:.2f}ms"
    else:
        return f"{t*1000000:.0f}us"

def header(text):
    return ui.label(text).style("font-size: 1.3em; font-weight: bold;")

def replace_long_brace_contents(code_str):
    def replacement(match):
        string = match.group(0)
        if len(string) > 300:
            # Extract the first 50 and last 30 characters from the string within brackets
            return '{' + string[1:51] + '...' + string[-31:-1] + '}'
        else:
            # If the string is not longer than 300 characters, return it unchanged
            return string

    # This regex matches sequences of characters wrapped in { }
    brace_content_regex = r'\{[\s\S]*?\}'

    # Use the sub method to replace the matched strings with the result of the replacement function
    return re.sub(brace_content_regex, replacement, code_str)


def code(c):
    c = replace_long_brace_contents(c)
    return ui.code(c).style("padding-right: 30px").classes("w-full")


@ui.refreshable
def details():
    if active_item:
        if active_item['event'] == "compilation":
            header(f"{active_item['source']['file']}: {active_item['source']['line']}")
            # with ui.row():
            code(active_item["source"]["block"] or active_item["emitted"])
            code(f"{active_item['emitted']}")
            header("IR")
            code(f"{active_item['task']}")
            header("Rewritten")
            code(f"{active_item['passes'][-1]['task']}")
            header("Passes")
            for p in active_item['passes']:
                with ui.column().classes("w-full"):
                    with ui.row():
                        ui.label(p['name'])
                        ui.label(f"({p['elapsed']*1000000:.0f} us)")
                    code(p['task'])
        elif active_item['event'] == "time":
            ui.label(f"{active_item['type']} | {format_time(active_item['elapsed'])} | {active_item['results']['count']}")
            vals = active_item['results']['values']
            if len(vals):
                keys = [k for k in vals[0].keys()]
                columns = [{'name': k, 'label': k, 'field': k, "align": "left"} for k in keys]
                ui.table(columns=columns, rows=vals)
            if "code" in active_item:
                header("Code")
                code(active_item["code"])
            # ui.label(f"{active_item['results']['values']}").style('white-space: pre;')

@ui.refreshable
def my_stuff():
    total_time = 0
    with ui.column():

        for (ix, obj) in enumerate(current_json_objects):
            me = None

            if obj['event'] == "compilation":
                me = code(obj["source"]["block"] or obj["emitted"])
                total_time += obj["emit_time"]
                for p in obj['passes']:
                    total_time += p['elapsed']
            elif obj['event'] == "time":
                total_time += obj['elapsed']
                with ui.row() as me:
                    me.style("padding: 5px 10px; border: 1px solid #544; border-radius: 5px; background:#322;")
                    ui.label(f"{obj['type']}")
                    ui.label(format_time(obj['elapsed']))
                    if obj.get('results'):
                        ui.label(f"{obj['results']['count']}")
            elif obj['event'] == "transaction_created":
                with ui.row() as me:
                    me.style("padding: 5px 10px; border: 1px solid #544; border-radius: 5px; background:#322;")
                    ui.label("transaction")
                    ui.label(obj["txn_id"])
            elif obj.get('type'):
                total_time += obj.get('elapsed', 0)
                with ui.row() as me:
                    me.style("padding: 5px 10px; border: 1px solid #544; border-radius: 5px; background:#322;")
                    ui.label(f"{obj['type']}")
                    ui.label(format_time(obj.get('elapsed', 0)))
                    if obj.get('results'):
                        ui.label(f"{obj['results']['count']}")
            elif 'span' not in obj['event'] and obj['event'] != "profile_event":
                with ui.row() as me:
                    me.style("padding: 5px 10px; border: 1px solid #544; border-radius: 5px; background:#322;")
                    ui.label(f"{obj['event']}")

            if me:
                me.classes("w-full")
                me.on("click", lambda ix=ix: set_item(ix))
                if ix != active_ix:
                    me.style('opacity: 0.5')
        with ui.row() as me:
            me.style("padding: 5px 10px; border: 1px solid #445; border-radius: 5px; background:#223; opacity: 0.5;")
            me.classes("w-full")
            if total_time > 1:
                ui.label(f"Total time: {total_time:.1f}s")
            else:
                ui.label(f"Total time: {total_time * 1000:.0f}ms")


def poll():
    global last_mod_time, active_item
    global current_json_objects
    # Check the last modification time of the file
    try:
        mod_time = os.path.getmtime('debug.jsonl')
        if last_mod_time is None or mod_time > last_mod_time:
            last_mod_time = mod_time
            # File has changed, read and parse the new content
            new_objects = []
            with open('debug.jsonl', 'r') as file:
                for line in file:
                    try:
                        # Parse each JSON object and add it to the list
                        obj = json.loads(line)
                        new_objects.append(obj)
                    except json.JSONDecodeError as e:
                        print(f"Error parsing JSON: {e}")
            # Update the current JSON objects
            current_json_objects = new_objects
            if active_ix is not None and len(current_json_objects) > active_ix:
                active_item = current_json_objects[active_ix]
            # Refresh the UI with the new objects
            my_stuff.refresh()
            details.refresh()
    except FileNotFoundError:
        pass

def main(host="0.0.0.0", port=8080):
    ui.dark_mode().enable()
    with ui.row():
        with ui.column() as c:
            c.style("cursor: pointer;")
            my_stuff()
        with ui.column() as c:
            c.style("padding-left: 2em;")
            details()

    ui.timer(1, poll)
    ui.run(reload=False, host=host, port=port)

if __name__ == "__main__":
    main()



# ---------------------------------------------------------------
# Test the RAI CLI commands
# ---------------------------------------------------------------

import sys
import pytest
from unittest.mock import patch
from click.testing import CliRunner
from relationalai.clients import config
from tests.unit.test_cli_mocks import (
    az_profile,
    sf_profile,
    mocked_engines_list,
    mocked_imports_status,
    mocked_imports_get_details,
    mocked_list_imports_response,
    mocked_import_stream_response
)

from relationalai.clients.config import ConfigStore

@pytest.fixture(autouse=True)
def run_before_each_test(mocker):
    # Mock get_config before each test so we can deal with the ensure_config
    get_config = mocker.patch('relationalai.tools.cli_helpers.get_config')
    gs_instance = get_config.return_value
    gs_instance.get.file_path = "some path"

    yield

    # Cleanup
    mocker.stopall()

# ---------------------------------------------------------------
# rai config:check
# ---------------------------------------------------------------
def test_config_check(mocker):
    from relationalai.tools.cli import config_check
    engine_response = mocked_engines_list[4]

    mocker.stopall()

    cs = mocker.patch.object(ConfigStore, 'get')
    cs.return_value = None

    gfc = mocker.patch('relationalai.clients.config._get_full_config')
    gfc.return_value = sf_profile, "some path"

    # ---------------------------------------------------------------
    #  Engine/Prop not found
    # ---------------------------------------------------------------

    runner = CliRunner()
    result = runner.invoke(config_check)

    assert "Error: Missing config value for 'engine'" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Engine not found
    # ---------------------------------------------------------------

    sf_profile['engine'] = 'foo'

    instance = mocker.patch('relationalai.tools.cli.get_resource_provider').return_value
    instance.get_engine.return_value = None

    result = runner.invoke(config_check)

    assert "Error: Configured engine 'foo' not found" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Engine in wrong state
    # ---------------------------------------------------------------

    instance.get_engine.return_value = engine_response
    instance.is_valid_engine_state.return_value = False

    result = runner.invoke(config_check)

    assert "Error: Engine 'foo' is in an invalid state: 'SUSPENDED'" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  All good
    # ---------------------------------------------------------------

    instance.is_valid_engine_state.return_value = True
    spy = mocker.spy(instance, 'get_engine')

    result = runner.invoke(config_check)

    spy.assert_called_once_with('foo')
    assert "Connection successful!" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Exception handling
    # ---------------------------------------------------------------

    with patch('relationalai.tools.cli.get_resource_provider', side_effect=Exception('Provider Error')):
      result = runner.invoke(config_check)

    assert "Error: Provider Error" in result.output
    assert result.exit_code == 0

# ---------------------------------------------------------------
# rai profiles:switch
# ---------------------------------------------------------------
def test_profiles_switch(mocker):
    from relationalai.tools.cli import profile_switch

    # Setup mocks for the ConfigStore used throughout the test
    rv = {'sf': { "platform":"snowflake" }, 'az': { "platform":"azure" }}
    mocker.patch('relationalai.clients.config.ConfigStore.get_profiles', return_value = rv)
    mocker.patch('relationalai.clients.config.ConfigStore.change_active_profile', return_value = True)
    mocker.patch('relationalai.clients.config.ConfigStore.save', return_value = True)

    # ---------------------------------------------------------------
    #  Switch profile (non-interactive mode)
    # ---------------------------------------------------------------

    spy_profile = mocker.spy(ConfigStore, 'change_active_profile')
    spy_save = mocker.spy(ConfigStore, 'save')

    runner = CliRunner()
    result = runner.invoke(profile_switch, ['--profile', 'az'])
    spy_profile.assert_called_once_with('az')
    spy_save.assert_called_once_with()
    assert "✓ Switched to profile 'az'" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Switch profile (interactive mode)
    # ---------------------------------------------------------------

    spy_profile = mocker.spy(ConfigStore, 'change_active_profile')
    spy_save = mocker.spy(ConfigStore, 'save')

    fuzzy_mock = mocker.patch('relationalai.tools.cli_controls.fuzzy')
    fuzzy_mock.return_value = 'sf'

    result = runner.invoke(profile_switch)

    spy_profile.assert_called_once_with('sf')
    spy_save.assert_called_once_with()
    assert "✓ Switched to profile 'sf'" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Clean slate - no profiles
    # ---------------------------------------------------------------

    # Kills all mocks - last in this test - position matters
    mocker.stopall()
    result = runner.invoke(profile_switch)
    assert 'No profiles found' in result.output
    assert result.exit_code == 0

# ---------------------------------------------------------------
# rai engines:get
# ---------------------------------------------------------------
def test_engines_get(mocker):
    from relationalai.tools.cli import engines_get
    mocked_response = mocked_engines_list[1]
    text_mock = mocker.patch('relationalai.tools.cli_controls.text')
    instance = mocker.patch('relationalai.tools.cli.get_resource_provider').return_value

    # ---------------------------------------------------------------
    #  Get the engine (non-interactive mode)
    # ---------------------------------------------------------------

    instance.get_engine.return_value = None

    runner = CliRunner()
    result = runner.invoke(engines_get, ['--name', 'baz'])
    assert 'Engine "baz" not found' in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Get the engine (interactive mode)
    # ---------------------------------------------------------------

    instance.get_engine.return_value = mocked_response

    # controls.text() is mocked to return "foo" here
    text_mock.return_value = "foo"

    result = runner.invoke(engines_get)
    assert "foo    XS     READY" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Engine not found
    # ---------------------------------------------------------------

    instance.get_engine.return_value = None

    result = runner.invoke(engines_get)
    assert 'Engine "foo" not found' in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Exception handling
    # ---------------------------------------------------------------

    with patch('relationalai.tools.cli.get_resource_provider', side_effect=Exception('Provider Error')):
      result = runner.invoke(engines_get)

    assert "Error fetching engine: Provider Error" in result.output
    assert result.exit_code == 1

# ---------------------------------------------------------------
# rai engines:create - Snowflake
# ---------------------------------------------------------------
def test_engines_create_snowflake(mocker):
    import relationalai.tools.cli
    import relationalai.tools.cli_helpers
    import relationalai.clients.config

    mocker.stopall()
    mocker.patch('relationalai.tools.cli.validate_engine_name', return_value=True)

    provider_instance = mocker.patch('relationalai.tools.cli.get_resource_provider').return_value

    spy_validate_engine_name = mocker.spy(relationalai.tools.cli, 'validate_engine_name')
    spy_create_engine_flow = mocker.spy(relationalai.tools.cli, 'create_engine_flow')
    spy_create_engine = mocker.spy(provider_instance, 'create_engine')

    cfg = config.Config(sf_profile)
    mocker.patch('relationalai.tools.cli.ensure_config', return_value=cfg)

    runner = CliRunner()
    result = runner.invoke(relationalai.tools.cli.engines_create, ['--name', 'foo', '--size', 'HighMem|S', '--pool', 'BAZ'])

    spy_create_engine_flow.assert_called_once_with(cfg, 'foo', 'HighMem|S', 'BAZ')
    spy_validate_engine_name.assert_called_once_with(cfg, 'foo')
    spy_create_engine.assert_called_once_with('foo', 'HighMem|S', 'BAZ')
    assert "Engine 'foo' created!" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Engine already exists
    # ---------------------------------------------------------------

    mocker.stopall()

    provider_instance = mocker.patch('relationalai.tools.cli_helpers.get_resource_provider').return_value
    provider_instance.get_engine.return_value = {'name': 'foo', 'size': 'HighMem|S', 'state': 'READY'}

    mocker.patch('relationalai.tools.cli.ensure_config', return_value=cfg)

    result = runner.invoke(relationalai.tools.cli.engines_create, ['--name', 'foo', '--size', 'HighMem|S', '--pool', 'BAZ'])
    assert "Engine 'foo' already exists" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Engine creation fails
    # ---------------------------------------------------------------

    mocker.stopall()

    provider_instance = mocker.patch('relationalai.tools.cli.get_resource_provider').return_value
    provider_instance.create_engine.side_effect = Exception('CREATE_ENGINE Error')

    mocker.patch('relationalai.tools.cli.validate_engine_name', return_value=True)
    mocker.patch('relationalai.tools.cli.ensure_config', return_value=cfg)

    result = runner.invoke(relationalai.tools.cli.engines_create, ['--name', 'foo', '--size', 'HighMem|S', '--pool', 'BAZ'])

    assert "Error: CREATE_ENGINE Error" in result.output
    assert result.exit_code == 0

# ---------------------------------------------------------------
# rai engines:create - Azure
# ---------------------------------------------------------------
def test_engines_create_azure(mocker):
    import relationalai.tools.cli
    import relationalai.clients.config

    mocker.stopall()
    mocker.patch('relationalai.tools.cli.validate_engine_name', return_value=True)

    provider_instance = mocker.patch('relationalai.tools.cli.get_resource_provider').return_value

    spy_validate_engine_name = mocker.spy(relationalai.tools.cli, 'validate_engine_name')
    spy_create_engine_flow = mocker.spy(relationalai.tools.cli, 'create_engine_flow')
    spy_create_engine = mocker.spy(provider_instance, 'create_engine')

    cfg = config.Config(az_profile)
    mocker.patch('relationalai.tools.cli.ensure_config', return_value=cfg)

    runner = CliRunner()
    result = runner.invoke(relationalai.tools.cli.engines_create, ['--name', 'baz', '--size', 'S'])

    spy_create_engine_flow.assert_called_with(cfg, 'baz', 'S', None)
    spy_validate_engine_name.assert_called_with(cfg, 'baz')
    spy_create_engine.assert_called_with('baz', 'S', '')
    assert "Engine 'baz' created!" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Engine already exists
    # ---------------------------------------------------------------

    mocker.stopall()

    provider_instance = mocker.patch('relationalai.tools.cli_helpers.get_resource_provider').return_value
    provider_instance.get_engine.return_value = {'name': 'baz', 'size': 'S', 'state': 'READY'}

    mocker.patch('relationalai.tools.cli.ensure_config', return_value=cfg)

    result = runner.invoke(relationalai.tools.cli.engines_create, ['--name', 'baz', '--size', 'S'])
    assert "Engine 'baz' already exists" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Engine creation fails
    # ---------------------------------------------------------------

    mocker.stopall()
    mocker.patch('relationalai.tools.cli.validate_engine_name', return_value=True)

    provider_instance = mocker.patch('relationalai.tools.cli.get_resource_provider').return_value
    provider_instance.create_engine.side_effect = Exception('CREATE_ENGINE Error')

    mocker.patch('relationalai.tools.cli.ensure_config', return_value=cfg)

    result = runner.invoke(relationalai.tools.cli.engines_create, ['--name', 'baz', '--size', 'S'])

    assert "Error: CREATE_ENGINE Error" in result.output
    assert result.exit_code == 0

# ---------------------------------------------------------------
# rai engines:delete
# ---------------------------------------------------------------
def test_engines_delete(mocker):
    from relationalai.tools.cli import engines_delete
    instance = mocker.patch('relationalai.tools.cli.get_resource_provider').return_value

    # ---------------------------------------------------------------
    #  Engine not found
    # ---------------------------------------------------------------

    instance.get_engine.return_value = None

    runner = CliRunner()
    result = runner.invoke(engines_delete, ['--name', 'baz'])
    assert "Engine 'baz' not found" in result.output
    assert result.exit_code == 1

    # ---------------------------------------------------------------
    #  Engine found - Delete engine
    # ---------------------------------------------------------------

    engine_response = {'name': 'foo', 'size': 'S', 'state': 'READY'}
    instance.get_engine.return_value = engine_response
    spy = mocker.spy(instance, 'delete_engine')

    result = runner.invoke(engines_delete, ['--name', 'foo'])
    spy.assert_called_once_with('foo', False)
    assert "Engine 'foo' deleted!" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  SETUP_CDC Exception when no --force flag
    # ---------------------------------------------------------------

    instance.get_engine.return_value = [{'name': 'bar', 'size': 'S', 'state': 'READY'}]
    instance.delete_engine.side_effect = Exception('SETUP_CDC Error')

    result = runner.invoke(engines_delete, ['--name', 'bar'])

    assert "Imports are setup to utilize this engine.\nUse 'rai engines:delete --force' to force delete engines.\nUse 'rai imports:setup --engine' to set a different engine for imports." in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  --force delete engine
    # ---------------------------------------------------------------

    instance.delete_engine.side_effect = None
    spy = mocker.spy(instance, 'delete_engine')

    result = runner.invoke(engines_delete, ['--name', 'bar', '--force'])

    spy.assert_called_once_with('bar', True)
    assert "Engine 'bar' deleted!" in result.output
    assert result.exit_code == 0

# ---------------------------------------------------------------
# rai engines:list
# ---------------------------------------------------------------
def test_engines_list(mocker):
    from relationalai.tools.cli import engines_list

    # ---------------------------------------------------------------
    #  Listing the engines
    # ---------------------------------------------------------------

    instance = mocker.patch('relationalai.tools.cli.get_resource_provider').return_value
    instance.list_engines.return_value = mocked_engines_list

    runner = CliRunner()
    result = runner.invoke(engines_list)

    assert "1   test   XS     READY" in result.output
    assert "2   foo    XS     READY" in result.output
    assert "3   bar    S      READY" in result.output
    assert "4   baz    M      PENDING" in result.output
    assert "5   goo    M      SUSPENDED" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Listing the engines (Filter by state)
    # ---------------------------------------------------------------

    instance = mocker.patch('relationalai.tools.cli.get_resource_provider').return_value
    spy = mocker.spy(instance, 'list_engines')

    result = runner.invoke(engines_list, ['--state', 'ready'])

    spy.assert_called_once_with('ready')
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  No engines found
    # ---------------------------------------------------------------

    instance.list_engines.return_value = []

    result = runner.invoke(engines_list)
    assert "No engines found" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Handle exception
    # ---------------------------------------------------------------

    with patch('relationalai.tools.cli.get_resource_provider', side_effect=Exception('Provider Error')):
      result = runner.invoke(engines_list)

    assert "Error fetching engines: Provider Error" in result.output
    assert result.exit_code == 1

# ---------------------------------------------------------------
# rai version
# ---------------------------------------------------------------
def test_rai_version(mocker):
    from relationalai.tools.cli import version
    from relationalai import __version__ as rai_version
    from railib import __version__ as railib_version

    runner = CliRunner()
    python_version = sys.version.split()[0]

    # ---------------------------------------------------------------
    #  No configuration file found with no latest version
    # ---------------------------------------------------------------

    instance = mocker.patch('relationalai.tools.cli.get_config').return_value
    instance.file_path = None

    latest_version = mocker.patch('relationalai.tools.cli.latest_version')
    latest_version.return_value = None

    result = runner.invoke(version)

    assert f"RelationalAI   {rai_version}" in result.output
    assert f"Rai-sdk        {railib_version}" in result.output
    assert f"Python         {python_version}" in result.output
    assert "App            No configuration file found. To create one, run: rai init" in result.output
    assert "→" not in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  No configuration file found with latest version
    # ---------------------------------------------------------------

    latest_version.return_value = "10000.1.1"

    result = runner.invoke(version)

    assert f"RelationalAI   {rai_version} → 10000.1.1" in result.output
    assert f"Rai-sdk        {railib_version} → 10000.1.1" in result.output
    assert f"Python         {python_version}" in result.output
    assert "App            No configuration file found. To create one, run: rai init" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Current is newer than latest
    # ---------------------------------------------------------------

    latest_version.return_value = "0.0.1"

    result = runner.invoke(version)
    assert f"RelationalAI   {rai_version}" in result.output
    assert "→ 0.0.1" not in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Configuration file found - Snowflake - Show App version
    # ---------------------------------------------------------------

    instance = mocker.patch('relationalai.tools.cli.get_config').return_value
    instance.file_path = "some path"
    instance.get.return_value = "snowflake"

    instance = mocker.patch('relationalai.tools.cli.get_resource_provider').return_value
    instance.get_version.return_value = "1.0.0"

    result = runner.invoke(version)

    assert f"RelationalAI   {rai_version}" in result.output
    assert f"Rai-sdk        {railib_version}" in result.output
    assert f"Python         {python_version}" in result.output
    assert "App            1.0.0" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Configuration file found - Azure
    # ---------------------------------------------------------------

    instance = mocker.patch('relationalai.tools.cli.get_config').return_value
    instance.file_path = "some path"
    instance.get.return_value = "azure"

    result = runner.invoke(version)

    assert f"RelationalAI   {rai_version}" in result.output
    assert f"Rai-sdk        {railib_version}" in result.output
    assert f"Python         {python_version}" in result.output
    assert "App" not in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Error getting versions - Azure/Snowflake
    # ---------------------------------------------------------------

    with patch('relationalai.tools.cli.get_config', side_effect=Exception('Config Error')):
      result = runner.invoke(version)

    assert "Error checking app version: Config Error" in result.output
    assert result.exit_code == 1

    # ---------------------------------------------------------------
    #  Error getting app version - Snowflake specific
    # ---------------------------------------------------------------

    instance = mocker.patch('relationalai.tools.cli.get_config').return_value
    instance.file_path = "some path"
    instance.get.return_value = "snowflake"

    with patch('relationalai.tools.cli.get_resource_provider', side_effect=Exception('Provider Error')):
      result = runner.invoke(version)

    assert f"RelationalAI   {rai_version}" in result.output
    assert f"Rai-sdk        {railib_version}" in result.output
    assert f"Python         {python_version}" in result.output
    assert "App   Provider Error" in result.output
    assert result.exit_code == 0

# ---------------------------------------------------------------
# rai imports:get
# ---------------------------------------------------------------
def test_imports_get(mocker):
    import relationalai.tools.cli
    import relationalai.clients.config

    mocker.stopall()

    mocker.patch.object(ConfigStore, 'get', return_value=None)
    mocker.patch('relationalai.clients.config._get_full_config', return_value=[sf_profile, "some path"])

    rp_instance = mocker.patch('relationalai.tools.cli.get_resource_provider').return_value
    rp_instance.list_imports.return_value = mocked_list_imports_response
    rp_instance.get_import_stream.return_value = mocked_import_stream_response

    # ---------------------------------------------------------------
    #  Interactive mode (no id provided)
    # ---------------------------------------------------------------
    mocker.patch('relationalai.tools.cli_controls.fuzzy', return_value='foo_id')

    runner = CliRunner()
    result = runner.invoke(relationalai.tools.cli.imports_get)

    assert "Imports fetched" in result.output
    assert "1   foo_id   FOO    FOO     2024-06…   JOE       1         LOADED" in result.output
    assert "2   bar_id   BAR    BAR     2024-01…   JAKE      1         LOADED" in result.output
    assert "3   baz_id   BAZ    BAZ     2024-03…   BLAKE     1         LOADED   Bad data" in result.output
    assert "Import details fetched" in result.output

    for detail in mocked_imports_get_details:
      assert detail in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Non-interactive mode (id provided)
    # ---------------------------------------------------------------
    result = runner.invoke(relationalai.tools.cli.imports_get, ['--id', 'foo_id'])

    assert "Imports fetched" in result.output
    assert "Import details fetched" in result.output

    for detail in mocked_imports_get_details:
      assert detail in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Non-interactive mode (provided id not found)
    # ---------------------------------------------------------------
    result = runner.invoke(relationalai.tools.cli.imports_get, ['--id', 'boo_id'])

    assert "Imports fetched" in result.output
    assert "Import 'boo_id' not found" in result.output
    assert result.exit_code == 0

# ---------------------------------------------------------------
# rai imports:list
# ---------------------------------------------------------------
def test_imports_list(mocker):
    import relationalai.tools.cli
    import relationalai.clients.config

    mocker.stopall()

    mocker.patch.object(ConfigStore, 'get', return_value=None)
    mocker.patch('relationalai.clients.config._get_full_config', return_value=[sf_profile, "some path"])

    rp_instance = mocker.patch('relationalai.tools.cli.get_resource_provider').return_value
    rp_instance.get_imports_status.return_value = mocked_imports_status
    rp_instance.list_imports.return_value = mocked_list_imports_response

    runner = CliRunner()

    # ---------------------------------------------------------------
    #  Happy path with imports configured
    # ---------------------------------------------------------------
    result = runner.invoke(relationalai.tools.cli.imports_list)

    assert "Imports config fetched" in result.output
    assert "Imports status: STARTED" in result.output
    assert "Imports fetched" in result.output
    assert "1   FOO    FOO     2024-06-03   JOE       1         LOADED" in result.output
    assert "2   BAR    BAR     2024-01-01   JAKE      1         LOADED" in result.output
    assert "3   BAZ    BAZ     2024-03-03   BLAKE     1         LOADED   Bad data" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Filter by id
    # ---------------------------------------------------------------
    spy = mocker.spy(rp_instance, 'list_imports')
    result = runner.invoke(relationalai.tools.cli.imports_list, ['--id', 'baz_id'])
    spy.assert_called_once_with('baz_id', None, None, None, None)
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Filter by name
    # ---------------------------------------------------------------
    spy = mocker.spy(rp_instance, 'list_imports')
    result = runner.invoke(relationalai.tools.cli.imports_list, ['--name', 'FOO'])
    spy.assert_called_once_with(None, 'FOO', None, None, None)
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Filter by model
    # ---------------------------------------------------------------
    spy = mocker.spy(rp_instance, 'list_imports')
    result = runner.invoke(relationalai.tools.cli.imports_list, ['--model', 'MODEL'])
    spy.assert_called_once_with(None, None, 'MODEL', None, None)
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Filter by status
    # ---------------------------------------------------------------
    spy = mocker.spy(rp_instance, 'list_imports')
    result = runner.invoke(relationalai.tools.cli.imports_list, ['--status', 'ACTIVE'])
    spy.assert_called_once_with(None, None, None, 'ACTIVE', None)
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Filter by creator
    # ---------------------------------------------------------------
    spy = mocker.spy(rp_instance, 'list_imports')
    result = runner.invoke(relationalai.tools.cli.imports_list, ['--creator', 'JOE'])
    spy.assert_called_once_with(None, None, None, None, 'JOE')
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Filter by all
    # ---------------------------------------------------------------
    spy = mocker.spy(rp_instance, 'list_imports')
    result = runner.invoke(relationalai.tools.cli.imports_list, [
      '--id', 'baz_id',
      '--name', 'FOO',
      '--model', 'MODEL',
      '--status', 'ACTIVE',
      '--creator', 'JOE'
    ])
    spy.assert_called_once_with('baz_id', 'FOO', 'MODEL', 'ACTIVE', 'JOE')
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Imports not configured - "SETUP_CDC" in exception
    # ---------------------------------------------------------------
    rp_instance.get_imports_status.side_effect = Exception('SETUP_CDC')
    result = runner.invoke(relationalai.tools.cli.imports_list)

    assert "Error: Imports are not configured." in result.output
    assert "To start use 'rai imports:setup' to set up an engine for imports." in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  No imports found
    # ---------------------------------------------------------------
    get_rp = mocker.patch('relationalai.tools.cli.get_resource_provider')
    rp_instance = get_rp.return_value
    rp_instance.list_imports.return_value = []
    result = runner.invoke(relationalai.tools.cli.imports_list)

    assert "Imports fetched" in result.output
    assert "No imports found" in result.output
    assert result.exit_code == 0

# ---------------------------------------------------------------
# rai imports:setup
# ---------------------------------------------------------------
def test_imports_setup(mocker):
    import relationalai.tools.cli
    import relationalai.clients.config

    mocker.stopall()

    mocker.patch.object(ConfigStore, 'get', return_value=None)
    mocker.patch('relationalai.clients.config._get_full_config', return_value=[sf_profile, "some path"])

    rp_instance = mocker.patch('relationalai.tools.cli.get_resource_provider').return_value
    rp_instance.get_imports_status.return_value = mocked_imports_status

    runner = CliRunner()

    # ---------------------------------------------------------------
    #  Display imports setup
    # ---------------------------------------------------------------

    result = runner.invoke(relationalai.tools.cli.imports_setup)

    assert "To suspend imports, use 'rai imports:setup --suspend'" in result.output
    assert "Field                 Value" in result.output
    assert "engine                FOO_ENG" in result.output
    assert "status                STARTED" in result.output
    assert "createdOn             2024-06-03 10:50:04" in result.output
    assert "lastSuspendedOn       N/A" in result.output
    assert "lastSuspendedReason   N/A" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Imports suspended
    # ---------------------------------------------------------------

    mocked_imports_status["status"] = 'suspended'
    mocked_imports_status["info"] = '{\n  "createdOn": "2024-06-03 10:50:04.735 -0700",\n  "lastSuspendedOn": "2024-06-13 10:50:04.735 -0700",\n  "lastSuspendedReason": "USER_SUSPENDED",\n  "state": "suspended"\n}'
    rp_instance.get_imports_status.return_value = mocked_imports_status
    result = runner.invoke(relationalai.tools.cli.imports_setup)

    assert "To resume imports, use 'rai imports:setup --resume'" in result.output
    assert "status                SUSPENDED" in result.output
    assert "createdOn             2024-06-03 10:50:04" in result.output
    assert "lastSuspendedOn       2024-06-13 10:50:04" in result.output
    assert "lastSuspendedReason   USER_SUSPENDED" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Suspend imports
    # ---------------------------------------------------------------

    spy_change_status = mocker.spy(rp_instance, 'change_imports_status')
    result = runner.invoke(relationalai.tools.cli.imports_setup, ['--suspend'])

    spy_change_status.assert_called_once_with(suspend=True)
    assert "Imports suspended" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Resume imports
    # ---------------------------------------------------------------

    spy_change_status = mocker.spy(rp_instance, 'change_imports_status')
    result = runner.invoke(relationalai.tools.cli.imports_setup, ['--resume'])

    spy_change_status.assert_called_once_with(suspend=False)
    assert "Imports resumed" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Setup with passed in engine
    # ---------------------------------------------------------------

    spy_set_engine = mocker.spy(relationalai.tools.cli, 'set_imports_engine')
    rp_instance.get_engine.return_value = mocked_engines_list[1]

    result = runner.invoke(relationalai.tools.cli.imports_setup, ['--engine', 'foo'])

    spy_set_engine.assert_called_once_with('foo')
    assert "Engine validated" in result.output
    assert "Imports engine set to 'foo'" in result.output
    assert result.exit_code == 0

    # ---------------------------------------------------------------
    #  Setup with invalid engine
    # ---------------------------------------------------------------

    rp_instance.get_engine.return_value = None
    result = runner.invoke(relationalai.tools.cli.imports_setup, ['--engine', 'foo'])

    assert "Engine 'foo' is invalid." in result.output
    assert "Please use 'rai engines:create' to create a valid engine." in result.output

    # ---------------------------------------------------------------
    #  Exception handling
    # ---------------------------------------------------------------

    rp_instance.get_engine.side_effect = Exception('GET_IMPORTS_STATUS Error')

    result = runner.invoke(relationalai.tools.cli.imports_setup, ['--engine', 'foo'])

    assert "Error: GET_IMPORTS_STATUS Error" in result.output
    assert result.exit_code == 1



#pyright: reportUnusedExpression=false

import relationalai as rai
from relationalai.std import rel, aggregates

model = rai.Model(name=globals().get("name", "test_agg_ordering"), config=globals().get("config"))
Foo = model.Type("Foo")
Bar = model.Type("Bar")

with model.rule():
    Foo.add(name="A", value=100)
    Foo.add(name="B", value=25)
    Foo.add(name="C", value=50)

#--------------------------------------------------
# Simple aggregate
#--------------------------------------------------

with model.query() as select:
    f = Foo()
    z = select(aggregates.count(f))

print(z.results)

#--------------------------------------------------
# Aggregate result used in calculation
#--------------------------------------------------

with model.query() as select:
    f = Foo()
    z = select(rel.minimum(2, aggregates.count(f)))

print(z.results)

#--------------------------------------------------
# Aggregate inputs indirectly referenced after
#--------------------------------------------------

with model.rule():
    b = Foo()
    aggregates.rank_desc(b.value) == 3
    b.set(Bar)

with model.query() as select:
    b = Bar()
    z = select(b.name)

print(z.results)

#--------------------------------------------------
# Ensure variables unify down through multiple aggs
#--------------------------------------------------

with model.rule():
    node = Foo()
    prop = node.value
    prop_rank = aggregates.rank_asc(prop)
    max_rank = aggregates.max(prop_rank)
    min_rank = aggregates.min(prop_rank)
    node.set(color=((prop_rank - min_rank) / (max_rank - min_rank)))

with model.query() as select:
    f = Foo()
    z = select(f.name, f.color)

print(z.results)





#pyright: reportUnusedExpression=false

# This tests that after we receive an error during query execution, that
# we still restore the temporary rule that captures things like Types
# being added to the model

import relationalai as rai

model = rai.Model(name=globals().get("name", "test_after_errors"), config=globals().get("config"))
model._config.set("compiler.use_value_types", True)

Person = model.Type("Person")
Adult = model.Type("Adult")

with model.rule():
    Person.add(name="Joe", age=74)
    Person.add(name="Bob", age=40)
    Person.add(name="Jane", age=10)

with model.rule():
    p = Person()
    p.age >= 18
    p.set(Adult)


try:
    with model.query() as select:
        a = Adult()
        a.foo > 0 # this will cause the rel to fail, since foo doesn't exist
        z = select(a, a.name, a.age)
except Exception:
    pass

# adding a type when value types are on causes initialization code to be added
# to the model's temporary rule
woop = model.Type("Woop")

with model.rule():
    p = Person()
    p.age >= 18
    p.set(Adult)

with model.query() as select:
    a = Adult()
    z = select(a, a.name, a.age)



import logging
import threading
import time
import traceback

from relationalai import debugging

PROFILE_POLL_SLEEP_S = 1.5

class TransactionEventsFeedbackHandler(logging.Handler):
    def __init__(self, resources):
        super().__init__()
        self.resources = resources
    
    def emit(self, record):
        if (isinstance(record.msg, dict) and
            record.msg["event"] == "span_start"):
            span = record.msg["span"]
            if span.type == "wait" and span.parent.type == "transaction":
                # spawn a thread to poll for profile events
                thread = threading.Thread(
                    target=self._poll_profile_events,
                    args=(record.msg["span"].attrs["txn_id"],),
                )
                thread.start()
                # wait for thread?
    
    def _poll_profile_events(self, txn_id):
        continuation_token = ''
        while True:
            try:
                resp = self.resources.get_transaction_events(txn_id, continuation_token)
                
                for event in resp['events']:
                    debugging.event('profile_event', txn_id=txn_id, profile_event=event)
                
                continuation_token = resp['continuation_token']
                # Empty continuation token indicates that we've reached the end of the stream.
                if continuation_token == '':
                    break
                time.sleep(PROFILE_POLL_SLEEP_S)
            except Exception as e:
                debugging.error(e)
                print('Error polling profile events:', e)
                traceback.print_exc()



#pyright: reportUnusedExpression=false
import relationalai as rai

model = rai.Model("MyCoolDatabase")
Person = model.Type("Person")
Adult = model.Type("Adult")

with model.rule():
    j=Person.add(name="Joe", age=74)
    b=Person.add(name="Bob", age=20)
    Person.add(name="Jane", age=10, friend=j).set(friend=b)

with model.rule():
    p = Person()
    p.age >= 18
    p.set(Adult)

with model.query() as select:
    a = Person()
    with model.found():
        a.friend.age > a.age + 30
    z = select(a, a.name, a.age, a.friend.name, a.friend.age)

print("People!\n")
print(z.results)


#pyright: reportUnusedExpression=false
from typing import Tuple
import relationalai as rai
from relationalai.clients.snowflake import Snowflake

model = rai.Model("MyCoolDatabase", dry_run=False)
sf = Snowflake(model)
Person = sf.sandbox.public.person
Address = sf.sandbox.public.address
Address.describe(
    personid=(Person, "person")
)

with model.rule():
    a = Address()
    a.personid > 0
    a.set(coolness=a.personid * 10)

with model.query() as select:
    a = Address()
    z = select(a.address, a.person.firstname, a.coolness)

@model.export("sandbox.public")
def my_analysis(personid: int) -> Tuple[str, int]:
    a = Address(
        person = Person(personid=personid)
    )
    return a.person.firstname, a.coolness

print("People!\n")
print(z.results)

print("\nCalling my analysis from Snowflake!\n")
for row in model.resources._exec("call sandbox.public.my_analysis(1);"):
    print(row)










from .clients import config as cfg
from . import clients
from . import dsl
from . import debugging
from . import metamodel
from . import rel
from .loaders import csv
from . import analysis
from . import tools
from . import dependencies
import importlib.metadata
from snowflake.connector import SnowflakeConnection
from .errors import RAIException

# Set up global exception handler for debugging
debugging.setup_exception_handler()
debugging.install_warninghook()

__version__ = importlib.metadata.version(__package__ or __name__)

def Model(
    name: str,
    *,
    profile: str | None = None,
    config: cfg.Config | None = None,
    dry_run: bool = False,
    debug: bool | None = None,
    connection: SnowflakeConnection | None = None,
    keep_model: bool = None,
    isolated: bool = None,
    nowait_durable: bool = None,
    format: str = "default",
):
    config = config or cfg.Config(profile=profile)
    if debug is None:
        debug = config.get("debug", False)
    if debug:
        from relationalai.tools.debugger_client import start_debugger_session
        start_debugger_session(config) # @TODO: add config / env vars / kwargs for debug_host and debug_port
    if not config.file_path:
        if cfg.legacy_config_exists():
            message = (
                "Use `rai init` to migrate your configuration file "
                "to the new format (raiconfig.toml)"
            )
        else:
            message = "No configuration file found. Please run `rai init` to create one."
        raise Exception(message)
    if config.get("platform") is None:
        config.set("platform", "snowflake")
    platform = config.get("platform")
    if platform != "snowflake" and connection is not None:
        raise ValueError("The `connection` parameter is only supported with the Snowflake platform")
    if dry_run is None:
        dry_run = config.get("compiler.dry_run", False)
    if keep_model is None:
        keep_model = config.get("model.keep", False)
    if isolated is None:
        isolated = config.get("model.isolated", True)
    if nowait_durable is None:
        nowait_durable = config.get("model.nowait_durable", True)

    try:
        if platform == "azure":
            model = clients.azure.Graph(
                name,
                profile=profile,
                config=config,
                dry_run=dry_run,
                isolated=isolated,
                keep_model=keep_model,
                format=format,
            )
        elif platform == "snowflake":
            model = clients.snowflake.Graph(
                name,
                profile=profile,
                config=config,
                dry_run=dry_run,
                isolated=isolated,
                connection=connection,
                keep_model=keep_model,
                nowait_durable=nowait_durable,
                format=format,
            )
        else:
            raise Exception(f"Unknown platform: {platform}")
    except RAIException as e:
        raise e.clone() from None
    model._client.exec_control(dependencies.generate_query(), lambda res: dependencies.check_dependencies(res, platform, config, name))
    return model

def Resources(profile:str|None=None, config:cfg.Config|None=None):
    config = config or cfg.Config(profile)
    platform = config.get("platform", "snowflake")
    if platform == "azure":
        return clients.azure.Resources(config=config)
    elif platform == "snowflake":
        return clients.snowflake.Resources(config=config)
    else:
        raise Exception(f"Unknown platform: {platform}")

def Graph(name:str, dry_run:bool=False):
    return Model(name, profile=None, dry_run=dry_run)

__all__ = ['Model', 'Resources', 'dsl', 'rel', 'debugging', 'metamodel', 'csv', 'analysis', 'tools']



import functools
import os
import sys
import traceback
from typing import Callable, TypeVar
from colorama import Style
from hypothesis import strategies as gen
from relationalai import metamodel as mm
import relationalai
from relationalai.clients.test import Install, Query
import gentest


T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")

GENTEST_ROOT_DIR = os.path.dirname(gentest.__file__)
PYREL_ROOT_DIR = os.path.dirname(relationalai.__file__)
PROJECT_DIR = os.path.dirname(os.path.dirname(PYREL_ROOT_DIR))

# Check if we're in a virtual environment
if hasattr(sys, 'real_prefix') or (hasattr(sys, 'base_prefix') and sys.base_prefix != sys.prefix):
    # We're in a virtual environment
    VENV_MODULE_DIR = os.path.join(sys.prefix, 'lib', f'python{sys.version_info.major}.{sys.version_info.minor}', 'site-packages')
else:
    # Not in a virtual environment
    VENV_MODULE_DIR = None

#===============================================================================
# Exception
#===============================================================================

def get_last_exception_entry(exception: Exception) -> tuple[str, int, str]:
    # Extract the traceback from the exception object
    tb = traceback.extract_tb(exception.__traceback__)
    # Get the last call before the exception was raised
    filename, lineno, _, text = tb[-1]
    return filename, lineno, text

def condense_traceback(exception: Exception):
    tb = traceback.extract_tb(exception.__traceback__)
    max_prefix_size = 0
    prefixes: list[str] = []
    linenos: list[int] = []
    messages: list[str] = []
    for filename, lineno, name, text in tb:
        prefix = f"{condense_path(filename)} in {name}"
        if prefix.startswith("_pytest") or prefix.startswith("pluggy"):
            continue

        if len(prefix) > max_prefix_size:
            max_prefix_size = len(prefix)
        prefixes.append(prefix)
        linenos.append(lineno)
        messages.append(text.strip())

    return "\n".join([f"{Style.DIM}{prefix: <{max_prefix_size}} {lineno: >4} |{Style.RESET_ALL} {msg}" for prefix, lineno, msg in zip(prefixes, linenos, messages)])

#===============================================================================
# String
#===============================================================================

UNDERLINE = "\033[4m"
END_UNDERLINE = "\033[0m"

def fmt_boxed(msg: str, style: str = "", box_style: str = ""):
    return f"{box_style}[{style} {msg} {box_style}]"

def fmt_header(start: str|None = None, end: str|None = None, demarc: str = "-", indent = 0, line_width = 80):
    used = 2 * indent
    parts: list[str] = ['  ' * indent, Style.DIM]
    if start:
        used += len(start) + 4 # for box
    if end:
        used += len(end) + 4 # for box

    if start:
        parts.append(fmt_boxed(start, Style.NORMAL, Style.DIM))
        parts.append(Style.NORMAL)

    interstitial = demarc * (line_width - used)
    parts.append(f"{Style.DIM}{interstitial}{Style.NORMAL}")
    if end:
        parts.append(fmt_boxed(end, Style.NORMAL, Style.DIM))
        parts.append(Style.NORMAL)

    return "".join(parts)

def fmt_h1(start: str|None = None, end: str|None = None, line_width = 80, indent = 0):
    return fmt_header(start, end, "#", line_width=line_width, indent=indent)

def fmt_h2(start: str|None = None, end: str|None = None, line_width = 80, indent = 1):
    return fmt_header(start, end, "=", indent=indent, line_width=line_width)

def fmt_h3(start: str|None = None, end: str|None = None, line_width = 80, indent = 2):
    return fmt_header(start, end, "-", indent=indent, line_width=line_width)

def fmt_indented(msg: str, indent = 0):
    if indent == 0:
        return msg

    pad = "  " * indent
    return "\n".join(f"{pad}{line}" for line in msg.splitlines())

def fmt_unindented(msg: str, indent: int|None = None):
    if indent is None:
        drop = len(msg) - len(msg.lstrip())
    else:
        drop = indent * 2

    if drop == 0:
        return msg

    return "\n".join(f"{line[drop:]}" for line in msg.splitlines())

def fmt_lines(text: str, prefix: str = ""):
    return "\n".join(
        f"{prefix}{Style.DIM}{ix + 1: >5} |{Style.RESET_ALL} {line}"
        for (ix, line) in enumerate(text.rstrip().splitlines()))

Decoration = tuple[str, str, str, str]
def decorate(gutter_pre: str = Style.DIM, gutter_post: str = Style.RESET_ALL, line_pre: str = "", line_post: str = ""):
    return (gutter_pre, gutter_post, line_pre, line_post)

DEFAULT_DECORATION = decorate()
def fmt_styled_lines(text: str, decorations: dict[int|None, Decoration], prefix: str = ""):
    def deco(ix: int, slot: int):
        return decorations.get(ix, DEFAULT_DECORATION)[slot]

    return "\n".join(
        f"{prefix}{deco(ix, 0)}{ix: >5} |{deco(ix, 1)}{deco(ix, 2)} {line}{deco(ix, 3)}"
        for (ix, line) in enumerate(text.rstrip().splitlines(), start=1))



def condense_path(filename: str):
    if VENV_MODULE_DIR and filename.startswith(VENV_MODULE_DIR):
        return os.path.relpath(filename, VENV_MODULE_DIR)
    if filename.startswith(PROJECT_DIR):
        return os.path.relpath(filename, PROJECT_DIR)
    return filename

#===============================================================================
# Dicts
#===============================================================================

def invert_dict(original: dict[K, V]) -> dict[V, list[K]]:
    inverted = {}
    for k, v in original.items():
        inverted.setdefault(v, []).append(k)

    return inverted

#===============================================================================
# Functions
#===============================================================================

def wraps(inner_fn: Callable):
    def apply_wrap(wrapper: Callable):
        wrapped = functools.wraps(inner_fn)(wrapper)
        for attr in dir(inner_fn):
            if not attr.startswith("__"):
                setattr(wrapped, attr, getattr(inner_fn, attr))

        return wrapped

    return apply_wrap




#===============================================================================
# Generation Strategies
#===============================================================================

@gen.composite
def gen_partitions(draw: gen.DrawFn, items: list[T]) -> tuple[dict[T, int], int]:
    shuffled = draw(gen.permutations(items))

    mapping: dict[T, int] = {}

    cursor = 0
    count = len(shuffled)
    group_ix = 0
    while cursor < count:
        size = draw(gen.integers(min_value=1, max_value=count - cursor))
        for ix in range(size):
            v = shuffled[cursor + ix]
            mapping[v] = group_ix
        cursor += size
        group_ix += 1

    return mapping, group_ix

#===============================================================================
# Metamodel
#===============================================================================

anon_printer = mm.Printer(unnamed_vars=True)
def fmt_task(task: mm.Task) -> str:
    anon_printer.indent = 0
    return anon_printer.print(task)


def stringify_blocks(blocks: list[Query|Install]) -> str:
    chunks: list[str] = []

    ix = 0
    for block in blocks:
        match block:
            case Install():
                if block.name != "pyrelstd":
                    name = block.name or "None"
                    chunks.append(name + "-"*(80 - len(name)))
                    chunks.append(fmt_task(block.task))
            case Query():
                name = f"query{ix}"
                chunks.append(name + "-"*(80 - len(name)))
                chunks.append(fmt_task(block.task))
                ix += 1

    return "\n".join(chunks)



from functools import reduce
import re
from typing import Any, Tuple, List
import base64

from pandas import DataFrame
import pandas as pd
import decimal

from .. import metamodel as m

#--------------------------------------------------
# Constants
#--------------------------------------------------

UNIXEPOCH = 62135683200000
MILLISECONDS_PER_DAY = 24 * 60 * 60 * 1000

def format_columns(result_frame:DataFrame, types:List[str]):
    for i, col in enumerate(result_frame.columns):
        if "UInt128" in types[i]:
            result_frame[col] = result_frame[col].apply(lambda x: base64.b64encode(x.tobytes()).decode()[:-2])
        elif "FixedDecimal" in types[i]:
            decimal_info = re.search(r"FixedDecimal\{Int(\d+), (\d+)\}", types[i])
            if decimal_info:
                bits = int(decimal_info.group(1))
                scale = int(decimal_info.group(2))
                if bits == 128:
                    result_frame[col] = result_frame[col].apply(lambda x: (decimal.Decimal(str((int(x[1]) << 64) + int(x[0]))) if x[1] > 0 else decimal.Decimal(str(x[0]))) / (10 ** scale))
                else:
                    result_frame[col] = result_frame[col].apply(lambda x: decimal.Decimal(str(x)) / (10 ** scale))
        elif "Int128" in types[i]:
            result_frame[col] = result_frame[col].apply(lambda x: (int(x[1]) << 64) + int(x[0]) if x[1] > 0 else x[0])
        elif "Missing" in types[i]:
            result_frame[col] = result_frame[col].apply(lambda x: None)
        elif types[i] == "Dates.DateTime":
            result_frame[col] = pd.to_datetime(result_frame[col] - UNIXEPOCH, unit="ms")
        elif types[i] == "Dates.Date":
            result_frame[col] = pd.to_datetime(result_frame[col] * MILLISECONDS_PER_DAY - UNIXEPOCH, unit="ms")
    return result_frame

def format_results(results, task:m.Task) -> Tuple[DataFrame, List[Any]]:
    data_frame = DataFrame()
    if len(results.results):
        has_cols = []
        for result in results.results:
            types = [t for t in result["relationId"].split("/") if t != "" and not t.startswith(":")]
            result_frame:DataFrame = result["table"].to_pandas()
            result_frame = format_columns(result_frame, types)

            result["table"] = result_frame
            if ":output/:cols" in result["relationId"]:
                col_ix = int(re.search(r":col([0-9]+)", result["relationId"]).group(1))
                key_cols = [f"id{i}" for i in range(0, len(result_frame.columns) - 1)]
                result_frame.columns = [*key_cols, f"v{col_ix}"]
                if col_ix < len(has_cols):
                    has_cols[col_ix] = pd.concat([has_cols[col_ix], result_frame], ignore_index=True)
                else:
                    has_cols.append(result_frame)
            elif ":output" in result["relationId"]:
                data_frame = pd.concat([data_frame, result_frame], ignore_index=True)
        if len(has_cols):
            key_cols = [f"id{i}" for i in range(0, len(has_cols[0].columns) - 1)]
            df_wide_reset = reduce(lambda left, right: pd.merge(left, right, on=key_cols, how='outer'), has_cols)
            data_frame = df_wide_reset.drop(columns=key_cols)
            data_frame.sort_values(by=[str(c) for c in data_frame.columns], ascending=[True] * len(data_frame.columns), inplace=True)
            data_frame = data_frame.reset_index(drop=True)

            # data_frame = df_wide
        ret_cols = task.return_cols() if task else []
        if len(ret_cols) and len(data_frame.columns) == len(ret_cols):
            data_frame.columns = task.return_cols()[0:len(data_frame.columns)]
    return (data_frame, results.problems)


import relationalai as rai
from relationalai.clients.snowflake import Snowflake, PrimaryKey
from relationalai.std import aggregates
from relationalai.std.graphs import Graph

# Get the model for the "power_transmission_network" schema.
model = rai.Model("power_transmission_network")

# Get the 'Snowflake' object for the model.
# You'll use this to access objects imported from the model's schema.
sf = Snowflake(model)

# Assign objects from the 'nodes' table to the 'NetworkNode' type
# and objects from the 'powerlines' table to the 'PowerLine' type.
# Tables are accessed as attributes of the 'sf' object via the
# pattern 'sf.<database_name>.<schema_name>.<table_name>'. Names
# are case-insensitive.
NetworkNode = sf.rai_getting_started.power_transmission_network.nodes
PowerLine = sf.rai_getting_started.power_transmission_network.powerlines

NetworkNode.describe(id=PrimaryKey)
PowerLine.describe(
    source_node_id=(NetworkNode, "source"),
    target_node_id=(NetworkNode, "target"),
)

# Define a type for nodes that need repair.
NeedsRepair = model.Type("NeedsRepair")

# Assign nodes with status "fail" to the NeedsRepair type.
with model.rule():
    node = NetworkNode()
    node.status == "fail"
    node.set(NeedsRepair)

# Loads critical to public safety and security are given higher priority.
with model.rule():
    node = NeedsRepair(type="load")
    with model.match():
        # Nodes with a description of "hospital", "military", or
        # "government" are given the highest priority.
        with node.description.in_(["hospital", "military", "government"]):
            node.set(load_priority=2)
        # "Load" nodes with other descriptions are given a lower priority.
        with model.case():
            node.set(load_priority=1)

# Get a new graph object. By default graphs are directed.
graph = Graph(model)

# Add 'NeedsRepair' nodes to the `graph.Node` type.
with model.rule():
    repair_node = NeedsRepair()
    graph.Node.add(repair_node)

# Add edges between `NeedsRepair` nodes that are connected by a 'PowerLine'.
with model.rule():
    line = PowerLine()
    graph.Edge.add(from_=NeedsRepair(line.source), to=NeedsRepair(line.target))

# Nodes upstream of critical loads are given higher priority,
# as are nodes with more downstream connections.
with model.rule():
    upstream, downstream = NeedsRepair(), NeedsRepair()
    graph.compute.is_reachable(upstream, downstream)
    upstream.set(
        load_priority=aggregates.max(downstream.load_priority.or_(0), per=[upstream]),
        connection_priority=aggregates.count(downstream, per=[upstream])
    )

# Prioritize nodes by ranking them in descendeing order first by load priority,
# then by connection priority, and finally by the node ID to break ties.
with model.rule():
    node = NeedsRepair()
    node.set(priority=aggregates.rank_desc(
        node.load_priority.or_(0),
        node.connection_priority.or_(0),
        node.id
    ))

# Query the model for the repair priority of each node.
with model.query() as select:
    node = NeedsRepair()
    node.priority <= 10  # Limit the results to the top 10 nodes.
    response = select(node.priority, node.id, node.type, node.description.or_(""))

print(response.results)
#     priority         type  description
# 0          1  transformer
# 1          2  transformer
# 2          3  transformer
# 3          4  transformer
# 4          5         load     hospital
# 5          6  transformer
# 6          7         load   commercial
# 7          8         load  residential
# 8          9  transformer
# 9         10  transformer

# Pass properties of 'NeedsRepair' objects to the graph so they can be
# displayed in the visualization.
with model.rule():
    repair = NeedsRepair()
    # Get the graph's 'Node' object for each 'NeedsRepair' object and
    # set the properties that will be displayed in the visualization.
    graph.Node(repair).set(
        id=repair.id,
        type=repair.type,
        description=repair.description,
        priority=repair.priority,
    )

# Visualie the graph. The 'visualize()' method accepts a style dictionary
# that lets you customize the appearance of nodes and edges in the graph.
graph.visualize(
    style={
        "node": {
            # Color load nodes red and all other nodes black.
            "color": lambda n: {"load": "red"}.get(n["type"], "black"),
            # Label nodes by their priority.
            "label": lambda n: n["priority"],
            # Include additional information when hovering over a node.
            "hover": lambda n: f"ID: {n['id']}\nType: {n['type']}\nDescription: {n.get('description', 'none')}",
        },
    },
).display()  # In Jupyter notebooks, .display() is not required.



import relationalai as rai
from relationalai import std

# Query snapshots are output in same order as these lists.
UNARY_ALGOS = [
    "min_degree",
    "max_degree",
    "avg_degree",
    "min_indegree",
    "max_indegree",
    "avg_indegree",
    "min_outdegree",
    "max_outdegree",
    "avg_outdegree",
]

BINARY_ALGOS = [
    "degree",
    "indegree",
    "outdegree",
]

model = rai.Model(name=globals().get("name", "test_degree"), config=globals().get("config"))
Object = model.Type("Object")
Relationship = model.Type("Relationship")

with model.rule():
    obj1 = Object.add(id=1)
    obj2 = Object.add(id=2)
    Relationship.add(from_=obj1, to=obj2)

graph = std.graphs.Graph(model)

with model.rule():
    r = Relationship()
    graph.Edge.add(r.from_, r.to)

for algo_name in UNARY_ALGOS:
    with model.query(tag=algo_name) as select:
        algo = getattr(graph.compute, algo_name)()
        select(algo)

for algo_name in BINARY_ALGOS:
    with model.query(tag=algo_name) as select:
        o = Object()
        algo = getattr(graph.compute, algo_name)(o)
        select(o, algo)



import re
import difflib
from colorama import Fore, init

# Initialize colorama
init(autoreset=True)

# def diff_to_str(a, b):
#     matcher = difflib.SequenceMatcher(None, a, b)
#     output = []
#     for opcode, a0, a1, b0, b1 in matcher.get_opcodes():
#         if opcode == 'equal':
#             output.append(a[a0:a1])
#         elif opcode == 'insert':
#             output.append(Fore.GREEN + b[b0:b1] + Fore.RESET)
#         elif opcode == 'delete':
#             output.append(Fore.RED + a[a0:a1] + Fore.RESET)
#         elif opcode == 'replace':
#             output.append(Fore.RED + a[a0:a1] + Fore.RESET)
#             output.append(Fore.GREEN + b[b0:b1] + Fore.RESET)
#     return ''.join(output)


def tokenize(s):
    # return re.findall(r'\w+|[^\w\s]|\s+', s)
    return re.findall(r'\S+|\s+', s)


def diff_to_str(a, b):
    # Tokenize the input strings while preserving whitespace
    tokens_a = tokenize(a)
    tokens_b = tokenize(b)

    # Create a SequenceMatcher to compare the token lists
    matcher = difflib.SequenceMatcher(None, tokens_a, tokens_b)

    output = []
    for opcode, a0, a1, b0, b1 in matcher.get_opcodes():
        if opcode == 'equal':
            output.extend(tokens_a[a0:a1])
        elif opcode == 'insert':
            output.extend([Fore.GREEN + token + Fore.RESET for token in tokens_b[b0:b1]])
        elif opcode == 'delete':
            output.extend([Fore.RED + token + Fore.RESET for token in tokens_a[a0:a1]])
        elif opcode == 'replace':
            output.extend([Fore.RED + token + Fore.RESET for token in tokens_a[a0:a1]])
            output.extend([Fore.GREEN + token + Fore.RESET for token in tokens_b[b0:b1]])
    return ''.join(output)



from typing import Any


class BuilderConsumedError(Exception):
    def __init__(self, builder: Any):
        super().__init__(f"{builder.__class__.__name__} modified after finish() called. finish() reuses the collections the builder populates, so this would mutate the resulting instance and break any generators dependant on it. Consider calling .snapshot() instead if the intent is to keep producing new instances from the same builder.")



from pathlib import Path

from relationalai.clients import config as cfg
from tests.util import all_files_in_dir, do_snapshot_test

@all_files_in_dir(Path(__file__).parent / "test_cases")
def test_snapshots(file_path: Path, snapshot, engine_config: cfg.Config, engine_size):
    do_snapshot_test(engine_config, engine_size, snapshot, file_path)



import datetime as dt
import numbers

from .. import dsl

# Custom types
_Date = dt.date | dsl.Producer
_DateTime = dt.datetime | dsl.Producer
_Integer = numbers.Integral | dsl.Producer
_String = str | dsl.Producer

# NOTE: Right now, common contains all Rel stdlib relations.
# If the stdlib is split into multiple namespaces, this will have to be updated.
_date_ns = dsl.global_ns.std.common

#--------------------------------------------------
# Types
#--------------------------------------------------

Date = _date_ns.Date
DateTime = _date_ns.DateTime


#--------------------------------------------------
# Constructors
#--------------------------------------------------

_make_date = getattr(_date_ns, "^Date")
_make_datetime = getattr(_date_ns, "^DateTime")

def _date(year: _Integer, month: _Integer, day: _Integer) -> dsl.Expression:
    return _make_date(year, month, day)

def _date_fromordinal(ordinal: _Integer) -> dsl.Expression:
    return _make_date(ordinal)

def _date_fromdatetime(datetime: _DateTime, tz: dt.tzinfo | _String | None = None) -> dsl.Expression:
    if tz is None:
        if isinstance(datetime, dt.datetime):
            tz = datetime.tzname() or "UTC"
        else:
            tz = "UTC"
    return _make_date(datetime, tz)

date = _date
date.fromordinal = _date_fromordinal
date.fromdatetime = _date_fromdatetime

def _datetime(
        year: _Integer,
        month: _Integer,
        day: _Integer,
        hour: _Integer = 0,
        minute: _Integer = 0,
        second: _Integer = 0,
        millisecond: _Integer = 0,
        tz: dt.tzinfo | _String = "UTC"
) -> dsl.Expression:
    if isinstance(tz, dt.tzinfo):
        tz = str(tz)
    return _make_datetime(year, month, day, hour, minute, second, millisecond, tz)

def _datetime_fromordinal(ordinal: _Integer) -> dsl.Expression:
    # Convert ordinal to milliseconds, since ordinals in Python are days but RAI DateTime expects milliseconds.
    ordinal_milliseconds = ordinal * 24 * 60 * 60 * 1000
    return _make_datetime(ordinal_milliseconds)

def _datetime_fromdate(
    date: _Date,
    hour: _Integer = 0,
    minute: _Integer = 0,
    second: _Integer = 0,
    millisecond: _Integer = 0,
    tz: _String = "UTC"
) -> dsl.Expression:
    return _make_datetime(date, hour, minute, second, millisecond, tz)

datetime = _datetime
datetime.fromordinal = _datetime_fromordinal
datetime.fromdate = _datetime_fromdate

#--------------------------------------------------
# Arithmetic
#--------------------------------------------------

def _date_dispatch(model: dsl.Graph, method: str, date: _Date | _DateTime, period: dsl.Producer) -> dsl.Expression:
    with model.match() as matched:
        with Date(date):
            matched.add(getattr(_date_ns,f"date_{method}")(date, period))
        with DateTime(date):
            matched.add(getattr(_date_ns, f"datetime_{method}")(date, period))
    return matched

def date_add(date: _Date | _DateTime, period: dsl.Producer) -> dsl.Expression:
    if isinstance(date, (dt.date, dt.datetime, dsl.Producer)):
        return _date_dispatch(dsl.get_graph(), "add", date, period)
    else:
        raise TypeError(f"date_add expects a date or datetime, got {type(date)}: {date!r}")

def date_subtract(date: _Date | _DateTime, period: dsl.Producer) -> dsl.Expression:
    if isinstance(date, (dt.date, dt.datetime, dsl.Producer)):
        return _date_dispatch(dsl.get_graph(), "subtract", date, period)
    else:
        raise TypeError(f"date_subtract expects a date or datetime, got {type(date)}: {date!r}")

#--------------------------------------------------
# Date Parts
#--------------------------------------------------

def year(date: _Date | _DateTime, tz: _String = "UTC") -> dsl.Expression:
    if isinstance(date, (dt.date, dt.datetime)):
        return date.year
    elif isinstance(date, dsl.Producer):
        model = dsl.get_graph()
        with model.match() as _year:
            with Date(date):
                _year.add(_date_ns.date_year(date))
            with DateTime(date):
                _year.add(_date_ns.datetime_year(date, tz))
        return _year
    else:
        raise TypeError(f"year expects a date or datetime, got {type(date)}: {date!r}")


def month(date: _Date | _DateTime, tz: _String = "UTC") -> dsl.Expression:
    if isinstance(date, (dt.date, dt.datetime)):
        return date.month
    elif isinstance(date, dsl.Producer):
        model = dsl.get_graph()
        with model.match() as _month:
            with Date(date):
                _month.add(_date_ns.date_month(date))
            with DateTime(date):
                _month.add(_date_ns.datetime_month(date, tz))
        return _month
    else:
        raise TypeError(f"month expects a date or datetime, got {type(date)}: {date!r}")


def day(date: _Date | _DateTime, tz: _String = "UTC") -> dsl.Expression:
    if isinstance(date, (dt.date, dt.datetime)):
        return date.day
    elif isinstance(date, dsl.Producer):
        model = dsl.get_graph()
        with model.match() as _day:
            with Date(date):
                _day.add(_date_ns.date_day(date))
            with DateTime(date):
                _day.add(_date_ns.datetime_day(date, tz))
        return _day
    else:
        raise TypeError(f"day expects a date or datetime, got {type(date)}: {date!r}")

def hour(date: _DateTime, tz: _String = "UTC") -> dsl.Expression:
    if isinstance(date, dt.datetime):
        return date.hour
    elif isinstance(date, dsl.Producer):
        return _date_ns.datetime_hour(date, tz)
    else:
        raise TypeError(f"hour expects a datetime, got {type(datetime)}: {datetime!r}")

def minute(date: _DateTime, tz: _String = "UTC") -> dsl.Expression:
    if isinstance(date, dt.datetime):
        return date.minute
    elif isinstance(date, dsl.Producer):
        return _date_ns.datetime_minute(date, tz)
    else:
        raise TypeError(f"minute expects a datetime, got {type(datetime)}: {datetime!r}")

def second(date: _DateTime, tz: _String = "UTC") -> dsl.Expression:
    if isinstance(date, dt.datetime):
        return date.second
    elif isinstance(date, dsl.Producer):
        return _date_ns.datetime_second(date)
    else:
        raise TypeError(f"second expects a datetime, got {type(datetime)}: {datetime!r}")

def week(date: _Date | _DateTime, tz: _String = "UTC") -> dsl.Expression:
    if isinstance(date, (dt.date, dt.datetime)):
        return date.isocalendar()[1]
    elif isinstance(date, dsl.Producer):
        model = dsl.get_graph()
        with model.match() as _week:
            with Date(date):
                _week.add(_date_ns.date_week(date))
            with DateTime(date):
                _week.add(_date_ns.datetime_week(date, tz))
        return _week
    else:
        raise TypeError(f"week expects a date or datetime, got {type(date)}: {date!r}")

def quarter(date: _Date | _DateTime, tz: _String = "UTC") -> dsl.Expression:
    if isinstance(date, (dt.date, dt.datetime)):
        return (date.month - 1) // 3 + 1
    elif isinstance(date, dsl.Producer):
        model = dsl.get_graph()
        with model.match() as _quarter:
            with Date(date):
                _quarter.add(_date_ns.date_quarterofyear(date))
            with DateTime(date):
                _quarter.add(_date_ns.datetime_quarterofyear(date, tz))
        return _quarter
    else:
        raise TypeError(f"quarter expects a date or datetime, got {type(date)}: {date!r}")

def weekday(date: _Date | _DateTime, tz: _String = "UTC") -> dsl.Expression:
    if isinstance(date, (dt.date, dt.datetime)):
        return date.weekday()
    elif isinstance(date, dsl.Producer):
        model = dsl.get_graph()
        with model.match() as _weekday:
            with Date(date):
                _weekday.add(_date_ns.date_dayofweek(date))
            with DateTime(date):
                _weekday.add(_date_ns.datetime_dayofweek(date, tz))
        return _weekday - 1  # Rel uses Monday=1, Python uses Monday=0
    else:
        raise TypeError(f"weekday expects a date or datetime, got {type(date)}: {date!r}")

def isoweekday(date: _Date | _DateTime, tz: _String = "UTC") -> dsl.Expression:
    if isinstance(date, (dt.date, dt.datetime)):
        return date.isoweekday()
    elif isinstance(date, dsl.Producer):
        model = dsl.get_graph()
        with model.match() as _isoweekday:
            with Date(date):
                _isoweekday.add(_date_ns.date_dayofweek(date))
            with DateTime(date):
                _isoweekday.add(_date_ns.datetime_dayofweek(date, tz))
        return _isoweekday
    else:
        raise TypeError(f"isoweekday expects a date or datetime, got {type(date)}: {date!r}")

def dayofyear(date: _Date | _DateTime, tz: _String = "UTC") -> dsl.Expression:
    if isinstance(date, (dt.date, dt.datetime)):
        return date.timetuple().tm_yday
    elif isinstance(date, dsl.Producer):
        model = dsl.get_graph()
        with model.match() as _yearday:
            with Date(date):
                _yearday.add(_date_ns.date_dayofyear(date))
            with DateTime(date):
                _yearday.add(_date_ns.datetime_dayofyear(date, tz))
        return _yearday
    else:
        raise TypeError(f"yearday expects a date or datetime, got {type(date)}: {date!r}")

#--------------------------------------------------
# Periods
#--------------------------------------------------

nanoseconds = getattr(_date_ns, "^Nanosecond")
microseconds = getattr(_date_ns, "^Microsecond")
milliseconds = getattr(_date_ns, "^Millisecond")
seconds = getattr(_date_ns, "^Second")
minutes = getattr(_date_ns, "^Minute")
hours = getattr(_date_ns, "^Hour")
days = getattr(_date_ns, "^Day")
weeks = getattr(_date_ns, "^Week")
months = getattr(_date_ns, "^Month")
years = getattr(_date_ns, "^Year")

def date_range(start: _Date | _DateTime | None = None, end: _Date | _DateTime | None = None, periods: int = 1, freq: str = "D") -> dsl.Expression:
    model = dsl.get_graph()
    period = {
        "us": microseconds,
        "ms": milliseconds,
        "s": seconds,
        "m": minutes,
        "H": hours,
        "D": days,
        "W": weeks,
        "M": months,
        "Y": years,
    }
    days2 = {
        "us": 86_400_000_000,
        "ms": 86_400_000,
        "s": 86_400,
        "m": 1440,
        "H": 24,
        "D": 1,
        "W": 1/7,
        "M": 1/(365/12),
        "Y": 1/365,
    }
    milliseconds2 = {
        "us": 1_000,
        "ms": 1,
        "s": 1/1_000,
        "m": 1/60_000,
        "H": 1/3_600_000,
        "D": 1/86_400_000,
        "W": 1/(86_400_000*7),
        "M": 1/(86_400_000*(365/12)),
        "Y": 1/(86_400_000*365),
    }
    date_func = date_add
    if start is None:
        start = end
        end = None
        date_func = date_subtract
    with model.match() as matched:
        with Date(start):
            if end is not None:
                num_days = dsl.rel.pyrel_dates_period_days(start, end)
                if freq in ["W", "M", "Y"]:
                    range_end = dsl.rel.float_int_convert(dsl.rel.ceil(num_days * days2[freq]))
                else:
                    range_end = num_days * days2[freq]
                ix = dsl.global_ns.std.common.range(0, range_end, 1)
            else:
                ix = dsl.global_ns.std.common.range(0, periods-1, 1)
            # if the frequency is less than a day, cast the dates to datetimes
            if freq in ["us", "ms", "s", "m", "H"]:
                _start = datetime.fromdate(start)
                if end is not None:
                    _end = datetime.fromdate(end)
            else:
                _start = start
                if end is not None:
                    _end = end
            _date = date_func(_start, period[freq](ix))
            if end is not None:
                _date <= _end
            matched.add(_date)
        with DateTime(start):
            if end is not None:
                num_ms = dsl.rel.pyrel_datetimes_period_milliseconds(start, end)
                if freq not in ["us", "ms"]:
                    _end = dsl.rel.float_int_convert(dsl.rel.ceil(num_ms * milliseconds2[freq]))
                else:
                    _end = num_ms * milliseconds2[freq]
                ix = dsl.global_ns.std.common.range(0, _end, 1)
            else:
                ix = dsl.global_ns.std.common.range(0, periods-1, 1)
            _date = date_func(start, period[freq](ix))
            if end is not None:
                _date <= end
            matched.add(_date)
    return matched

#--------------------------------------------------
# Exports
#--------------------------------------------------

__all__ = [
    "Date",
    "date",
    "DateTime",
    "datetime",
    "date_add",
    "date_subtract",
    "year",
    "month",
    "day",
    "hour",
    "minute",
    "second",
    "week",
    "quarter",
    "weekday",
    "isoweekday",
    "dayofyear",
    "nanoseconds",
    "microseconds",
    "milliseconds",
    "seconds",
    "minutes",
    "hours",
    "days",
    "weeks",
    "months",
    "years",
    "date_range",
]


import pytest
from relationalai import dependencies
from relationalai.dependencies import version_range
from relationalai.errors import RAIInvalidVersionWarning

#
# Dependencies tests
#

def test_compare_dependencies():
    expected = [
        ('std', '24c1872c-de8a-5b12-7648-3d72f007a7a9', version_range('0.1.0', '0.2.0')),
        ('graphlib', '04cd9c07-16aa-4433-bdd7-adfdfaee317d', version_range('0.1.0', '0.2.0'))
    ]
    # correct (does not raise exception or warning)
    dependencies._compare_dependencies(
        expected,
        {
            ('std', '24c1872c-de8a-5b12-7648-3d72f007a7a9'): '0.1.0',
            ('graphlib', '04cd9c07-16aa-4433-bdd7-adfdfaee317d'): '0.1.2'
        },
        None,
        platform="snowflake",
        config=None,
        model_name=""
    )

    # missing a library
    with pytest.warns(RAIInvalidVersionWarning):
        dependencies._compare_dependencies(
            expected,
            {
                ('std', '24c1872c-de8a-5b12-7648-3d72f007a7a9'): '0.1.0'
            },
            None,
            platform="snowflake",
            config=None,
            model_name=""
        )

    # library version is too old
    with pytest.warns(RAIInvalidVersionWarning):
        dependencies._compare_dependencies(
            expected,
            {
                ('std', '24c1872c-de8a-5b12-7648-3d72f007a7a9'): '0.0.23',
                ('graphlib', '04cd9c07-16aa-4433-bdd7-adfdfaee317d'): '0.1.2'
            },
            None,
            platform="snowflake",
            config=None,
            model_name=""
        )

    # library version is too new
    with pytest.warns(RAIInvalidVersionWarning):
        dependencies._compare_dependencies(
            expected,
            {
                ('std', '24c1872c-de8a-5b12-7648-3d72f007a7a9'): '0.2.0',
                ('graphlib', '04cd9c07-16aa-4433-bdd7-adfdfaee317d'): '0.1.2'
            },
            None,
            platform="snowflake",
            config=None,
            model_name=""
        )

    # database needs an update because there's no version data and no std version. This is
    # a reflection of old databases, before we started versioning libraries.
    with pytest.warns(RAIInvalidVersionWarning):
        dependencies._compare_dependencies(expected, {}, None, platform="snowflake", config=None, model_name="")

    # relationalai library needs an update because there's a version for std but there is
    # nothing in static lock. This is future proofing when we move versions to some other
    # place, and this version of the library is still in the wild
    with pytest.warns(RAIInvalidVersionWarning):
        dependencies._compare_dependencies(expected, {}, '0.1.0', platform="snowflake", config=None, model_name="")

    # currently not possible, but for completeness, a case where one library is too low and
    # another one is too high, which gives an "Incompatible" warning
    with pytest.warns(RAIInvalidVersionWarning):
        dependencies._compare_dependencies(
            expected,
            {
                ('std', '24c1872c-de8a-5b12-7648-3d72f007a7a9'): '0.0.1',
                ('graphlib', '04cd9c07-16aa-4433-bdd7-adfdfaee317d'): '2.1.2'
            },
            '0.1.0',
            platform="snowflake",
            config=None,
            model_name=""
        )

def test_generate_query():
    assert dependencies._generate_query([
        ('std', '24c1872c-de8a-5b12-7648-3d72f007a7a9', 'irrelevant'),
        ('graphlib', '04cd9c07-16aa-4433-bdd7-adfdfaee317d', 'irrelevant')
    ]).strip() == '''
    @no_diagnostics(:UNDEFINED_IDENTIFIER)
    def output[:std]: { std::version }

    @no_diagnostics(:TYPE_MISMATCH)
    def output(:static_lock, name, uuid, version):
        rel(:pkg, :std, :pkg, :project, :static_lock, name, uuid, version) and
        {
            ("std", "24c1872c-de8a-5b12-7648-3d72f007a7a9") ;
            ("graphlib", "04cd9c07-16aa-4433-bdd7-adfdfaee317d")
        }(name, uuid)'''.strip()



from typing import TypeVar, Callable, Optional
from hypothesis import strategies as gen
from hypothesis.strategies import SearchStrategy as Strategy

T = TypeVar('T', bound=tuple)
def staged(apply: Callable[..., Strategy[T]], initial: T, until: Optional[Callable[..., bool]] = None, max_stages: int = 5):
    if max_stages <= 0 or until and until(*initial):
        return gen.just(initial)

    return apply(*initial).flatmap(lambda v: staged(apply, v, until=until, max_stages=max_stages - 1))



from __future__ import annotations
from enum import Enum
import inspect
from itertools import zip_longest
import re
import threading
import typing
from typing import Any, Dict, List, Optional, Set, Tuple, TypeGuard, Union, get_type_hints
import numbers
import os
import sys
import datetime
import hashlib
import traceback
import rich

from pandas import DataFrame

from .clients.client import Client

from .metamodel import Behavior, Builtins, ActionType, Var, Task, Action, Builder, Type as mType, Property as mProperty
from . import debugging
from .errors import Errors, InvalidPropertySetException, NonCallablePropertyException, OutOfContextException, RAIException, RelQueryError, ReservedPropertyException, VariableOutOfContextException

from bytecode import Instr, Bytecode

#--------------------------------------------------
# Constants
#--------------------------------------------------

RESERVED_PROPS = ["add", "set", "persist", "unpersist"]

#--------------------------------------------------
# Helpers
#--------------------------------------------------

def to_var(x:Any):
    if isinstance(x, Var):
        return x
    if getattr(x, "_to_var", None):
        return to_var(x._to_var())
    if isinstance(x, ContextSelect):
        return x._vars[0]
    if isinstance(x, mProperty):
        return Var(value=x)
    if isinstance(x, mType):
        return Var(value=x)
    if isinstance(x, Producer):
        x._use_var()
        return x._var
    if isinstance(x, list) or isinstance(x, tuple):
        return Var(Builtins.Any, value=[to_var(i) for i in x])
    if isinstance(x, str):
        return Var(Builtins.String, None, x)
    if isinstance(x, numbers.Number):
        return Var(Builtins.Number, None, x)
    if isinstance(x, datetime.datetime) or isinstance(x, datetime.date):
        return Var(value=x)
    raise Exception(f"Unknown type: {type(x)}\n{x}")

build = Builder(to_var)

def to_list(x:Any):
    if isinstance(x, list):
        return x
    if isinstance(x, tuple):
        return list(x)
    return [x]

def is_static(x:Any):
    if isinstance(x, Var):
        return x.value is not None
    if isinstance(x, Type):
        return True
    if isinstance(x, str):
        return True
    if isinstance(x, Producer):
        return is_static(to_var(x))
    if isinstance(x, numbers.Number):
        return True
    if isinstance(x, list):
        return all(is_static(i) for i in x)
    if isinstance(x, tuple):
        return all(is_static(i) for i in x)
    if isinstance(x, dict):
        return all(is_static(i) for i in x.values())
    return False

def is_collection(x:Any) -> TypeGuard[Union[List, Tuple, Set]]:
    return isinstance(x, list) or isinstance(x, tuple) or isinstance(x, set)

def rel_dict(**kwargs):
    return InlineRelation(get_graph(), [
        (Symbol(k), v) for k, v in kwargs.items()
    ])

#--------------------------------------------------
# Base
#--------------------------------------------------

id = 0
def next_id():
    global id
    id += 1
    return id

#--------------------------------------------------
# Producer
#--------------------------------------------------

class Producer():
    def __init__(self, graph:'Graph'):
        self._id = next_id()
        self._graph = graph
        self._subs = {}

    def __getattr__(self, name: str) -> Any:
        if name.startswith("_"):
            return None
        if name == "getdoc":
            rich.print("[red bold]GETDOC CALLED")
            traceback.print_stack()
            return
        self._subs[name] = self._make_sub(name, self._subs.get(name))
        return self._subs[name]

    def _make_sub(self, name:str, existing:Optional['Producer']=None) -> Any:
        raise Exception("Implement Producer._make_sub")

    def _use_var(self):
        pass

    #--------------------------------------------------
    # Boolean overloads
    #--------------------------------------------------

    def __bool__(self):
        # This doesn't seem to be safe as Python can call bool on Producers in lots of random cases
        # InvalidBoolWarning(Errors.call_source())
        # class_name = self.__class__.__name__
        # TODO: Refactor this to be Exception instead of warning, so we could directly raise it if we bring back this
        # raise TypeError(f"Can't convert an {class_name} to a boolean.")
        return True

    #--------------------------------------------------
    # Math overloads
    #--------------------------------------------------

    def _wrapped_op(self, op, left, right):
        args = [left, right]
        return Expression(self._graph, op, args)

    def __add__(self, other):
        return self._wrapped_op(Builtins.plus, self, other)
    def __radd__(self, other):
        return self._wrapped_op(Builtins.plus, other, self)

    def __mul__(self, other):
        return self._wrapped_op(Builtins.mult, self, other)
    def __rmul__(self, other):
        return self._wrapped_op(Builtins.mult, other, self)

    def __sub__(self, other):
        return self._wrapped_op(Builtins.minus, self, other)
    def __rsub__(self, other):
        return self._wrapped_op(Builtins.minus, other, self)

    def __truediv__(self, other):
        return self._wrapped_op(Builtins.div, self, other)
    def __rtruediv__(self, other):
        return self._wrapped_op(Builtins.div, other, self)

    def __floordiv__(self, other):
        return self._wrapped_op(Builtins.floor_div, self, other)
    def __rfloordiv__(self, other):
        return self._wrapped_op(Builtins.floor_div, other, self)

    def __pow__(self, other):
        return self._wrapped_op(Builtins.pow, self, other)
    def __rpow__(self, other):
        return self._wrapped_op(Builtins.pow, other, self)

    def __mod__(self, other):
        return self._wrapped_op(Builtins.mod, self, other)
    def __rmod__(self, other):
        return self._wrapped_op(Builtins.mod, other, self)

    def __neg__(self):
        return self._wrapped_op(Builtins.mult, self, -1)

    #--------------------------------------------------
    # Filter overloads
    #--------------------------------------------------

    def __gt__(self, other):
        return self._wrapped_op(Builtins.gt, self, other)
    def __ge__(self, other):
        return self._wrapped_op(Builtins.gte, self, other)
    def __lt__(self, other):
        return self._wrapped_op(Builtins.lt, self, other)
    def __le__(self, other):
        return self._wrapped_op(Builtins.lte, self, other)
    def __eq__(self, other) -> Any:
        return self._wrapped_op(Builtins.eq, self, other)
    def __ne__(self, other) -> Any:
        return self._wrapped_op(Builtins.neq, self, other)

    #--------------------------------------------------
    # Context management
    #--------------------------------------------------

    def __enter__(self):
        self._graph._push(self)

    def __exit__(self, *args):
        self._graph._pop(self)

#--------------------------------------------------
# Context
#--------------------------------------------------

class TaskExecType(Enum):
    Query = 1
    Rule = 2
    Procedure = 3
    Export = 4

class ContextSelect(Producer):
    def __init__(self, context:'Context'):
        super().__init__(context.graph)
        self._context = context
        self._select_len = None
        self._insts = []
        self._vars = []
        self._props = {}

    def _assign_vars(self):
        task = self._context._task
        if not len(self._vars) and self._select_len:
            self._insts = to_list(Vars(self._select_len))
            self._vars = [to_var(v) for v in self._insts]
            task.properties = [Builtins.Relation.properties[i] for i in range(self._select_len)]
            task.bindings.update({Builtins.Relation.properties[i]: v for i, v in enumerate(self._vars)})

    def __getattr__(self, name: str) -> Any:
        if name.startswith("_"):
            return None
        if name in self._props:
            return Instance(self._context.graph, ActionType.Get, [], {}, var=to_var(self._props[name]))
        else:
            return getattr(Instance(self._context.graph, ActionType.Get, [], {}, var=to_var(self._vars[0])), name)

    #--------------------------------------------------
    # Return handling
    #--------------------------------------------------

    def _do_return(self, args, distinct=False):
        graph = self._context.graph
        task = self._context._task
        if task.behavior == Behavior.Query \
            and self._context._exec_type in [TaskExecType.Query, TaskExecType.Procedure]:
            if isinstance(args[0], tuple):
                args = args[0]
            export = self._context._format == "snowpark"
            graph._action(build.return_(list(args), export=export, distinct=distinct))
        else:
            #TODO: good error message depending on the type of task we're dealing with
            raise Exception("Can't select in a non-query")
        return self._context

    def __call__(self, *args: Any) -> Any:
        return self._do_return(args)

    def distinct(self, *args:Any) -> Any:
        return self._do_return(args, distinct=True)

    #--------------------------------------------------
    # Add for `model.match() as m`
    #--------------------------------------------------

    def add(self, item, **kwargs):
        arg_len = len(kwargs) + 1
        if self._select_len is not None and arg_len != self._select_len:
            raise Exception("Add must be provided the same arguments in each branch")
        self._select_len = arg_len
        self._assign_vars()
        if len(self._props) and set(self._props.keys()) != set(kwargs.keys()):
            raise Exception("Add must be provided the same properties in each branch")
        elif len(self._props) == 0:
            for k, v in zip(kwargs.keys(), self._vars[1:]):
                v.name = k
                self._props[k] = v

        graph = self._context.graph
        graph._action(build.return_([item, *[kwargs[k] for k in self._props.keys()]]))

class Context():
    def __init__(self, graph:'Graph', *args, behavior=Behavior.Query, op=None,
                 exec_type=TaskExecType.Rule, dynamic=False, format="default",
                 props=None, engine=None, tag=None):
        self._id = next_id()
        self.results = DataFrame()
        self.graph = graph
        self._task = Task(behavior=behavior)
        self._op = op
        self._args = list(args)
        self._exec_type = exec_type
        self._select_len = None
        self._dynamic = dynamic
        self._props = props or {}
        self._engine= engine
        self._tag = tag # for benchmark reporting
        self._format = self._resolve_format(format)

    def _resolve_format(self, format):
        if format == "default":
            if self.graph._format != "default":
                return self.graph._format
            if self.graph._config.get("platform") == "snowflake":
                # In the future we may want to default to snowpark
                return "pandas"
            return "pandas"
        return format

    def __enter__(self):
        debugging.set_source(self._task, dynamic=self._dynamic)
        self.graph._push(self)
        return ContextSelect(self)

    def __exit__(self, *args):
        # If no exception info has been passed to args,
        # then proceed with the normal exit process.
        # Otherwise, return False to propagate the exception.
        if args[0] is None:
            try:
                self.graph._pop(self)
            except KeyboardInterrupt as e:
                print("Canceling transactions...")
                self.graph.resources.cancel_pending_transactions()
                raise e
            except RAIException as e:
                raise e.clone() from None
            except RelQueryError as e:
                raise e.clone() from None
        else:
            self.graph._pop(self, exec=False)
            if isinstance(args[1], RAIException):
                raise args[1].clone() from None
        return False

    def __iter__(self):
        if self._exec_type != TaskExecType.Query:
            raise Exception("Can't iterate over a non-query task")
        else:
            return self.results.itertuples(index=False)

    def _repr_html_(self):
        if self._exec_type == TaskExecType.Query:
            return self.results.to_html(index=False)

    def __str__(self):
        if self._exec_type == TaskExecType.Query:
            return self.results.to_string(index=False)
        return super().__str__()

#--------------------------------------------------
# Type
#--------------------------------------------------

def hash_values_sha256_truncated(args):
    combined = ''.join(map(str, args))
    combined_bytes = combined.encode('utf-8')
    hasher = hashlib.sha256()
    hasher.update(combined_bytes)
    hash_128_bit = hasher.digest()[:16]
    return hash_128_bit

class Type(Producer):
    def __init__(self, graph:'Graph', name:str, builtins:List[str] = [], scope:str=""):
        super().__init__(graph)
        self._type = mType(scope+name)
        self._scope = scope
        if graph._config.get("compiler.use_value_types", False):
            self._type.parents.append(Builtins.ValueType)
            install = build.install(self._type)
            self._graph._action(install)
            debugging.set_source(install)

    def __call__(self, *args, **kwargs):
        return Instance(self._graph, ActionType.Get, [self, *args], kwargs, name=self._type.name.lower(), scope=self._scope)

    def add(self, *args, **kwargs):
        inst = Instance(self._graph, ActionType.Bind, [self, *args], kwargs, name=self._type.name.lower(), is_add=True, scope=self._scope)
        if is_static(args) and is_static(kwargs):
            params = [Var(value=t.name) for t in inst._action.types]
            params.extend(inst._action.bindings.values())
            inst._action.entity.value = hash_values_sha256_truncated(params)
        elif all([isinstance(a, Type) for a in args]):
            self._graph._action(build.ident(inst._action))
        inst._add_to_graph()
        return inst

    def persist(self, *args, **kwargs):
        inst = Instance(self._graph, ActionType.Persist, [self, *args], kwargs, name=self._type.name.lower(), is_add=True, scope=self._scope)
        if is_static(args) and is_static(kwargs):
            params = [Var(value=t.name) for t in inst._action.types]
            params.extend(inst._action.bindings.values())
            inst._action.entity.value = hash_values_sha256_truncated(params)
        elif all([isinstance(a, Type) for a in args]):
            self._graph._action(build.ident(inst._action))
        inst._add_to_graph()
        return inst

    def extend(self, *args, **kwargs):
        for arg in args:
            if not isinstance(arg, Type):
                raise Exception("Can only extend a type with another type")
            with self._graph.rule(dynamic=True):
                a = arg()
                a.set(self)
            with self._graph.rule(dynamic=True):
                a = arg()
                neue = self(a)
                for k, v in kwargs.items():
                    if isinstance(v, Property):
                        v = getattr(a, v._prop.name)
                    neue.set(**{k:v})

    def define(self, **kwargs):
        for k, v in kwargs.items():
            if isinstance(v, tuple):
                (other_type, left, right) = v
                with self._graph.rule():
                    inst = other_type()
                    me = self()
                    getattr(inst, right) == getattr(me, left) # type: ignore
                    if getattr(self, k).is_multi_valued:
                        getattr(me, k).add(inst)
                    else:
                        me.set(**{k: inst})
            else:
                raise Exception("Define requires a tuple of (Type, left, right)")

    def __or__(self, __value: Any) -> 'TypeUnion':
        if isinstance(__value, Type):
            return TypeUnion(self._graph, [self, __value])
        if isinstance(__value, TypeUnion):
            return TypeUnion(self._graph, [self, *__value._types])
        raise Exception("Can't or a type with a non-type")

    def _make_sub(self, name: str, existing=None):
        if existing is not None:
            return existing
        return Property(self._graph, name, [self._type], self, scope=self._scope)

    def known_properties(self):
        return [p.name.removeprefix(self._scope) for p in self._type.properties]

#--------------------------------------------------
# TypeUnion
#--------------------------------------------------

class TypeUnion(Producer):
    def __init__(self, graph:'Graph', types:List[Type]):
        super().__init__(graph)
        self._types = types

    def __call__(self, *args, **kwargs) -> 'ContextSelect':
        if not len(self._graph._stack.stack):
            raise Exception("Can't create an instance outside of a context")
        graph = self._graph
        with graph.union(dynamic=True) as union:
            for t in self._types:
                with graph.scope():
                    union.add(t(*args, **kwargs))
        return union

    def __or__(self, __value: Any) -> 'TypeUnion':
        if isinstance(__value, Type):
            return TypeUnion(self._graph, [*self._types, __value])
        if isinstance(__value, TypeUnion):
            return TypeUnion(self._graph, [*self._types, *__value._types])
        raise Exception("Can't or a type with a non-type")

    def _make_sub(self, name: str, existing=None):
        if existing is not None:
            return existing
        return Property(self._graph, name, [t._type for t in self._types], self)

#--------------------------------------------------
# Property
#--------------------------------------------------

class Property(Producer):
    def __init__(self, graph:'Graph', name:str, types:List[mType], provider:Type|TypeUnion, scope:str=""):
        super().__init__(graph)
        self._name = name
        self._type = types[0]
        self._scope = scope
        self._provider = provider
        self._prop = build.property_named(scope+name, types)

    def __call__(self, key:Any, value:Any):
        action = build.relation_action(ActionType.Get, self._prop, [key, value])
        self._graph._action(action)

    def _use_var(self):
        raise Exception("Support properties being used as vars")

    def _make_sub(self, name: str, existing=None):
        raise Exception("Support properties on properties?")

    def to_property(self):
        return self._prop

    def has_many(self):
        self._graph._check_property(self._prop, multi_valued=True)

    @property
    def is_multi_valued(self):
        return self._graph._prop_is_multi.get(self._name)

#--------------------------------------------------
# Instance
#--------------------------------------------------

class Instance(Producer):
    def __init__(self, graph:'Graph', action_type:ActionType, positionals:List[Any], named:Dict[str,Any], var:Var|None=None, name=None, is_add=False, scope:str=""):
        super().__init__(graph)
        self._action = Action(action_type, to_var(var) if var else Var(name=name))
        self._actions = [self._action]
        self._sets = {}
        self._context = graph._stack.active()
        self._scope = scope
        available_types = []
        last_pos_var = None

        #--------------------------------------------------
        # Positionals
        #--------------------------------------------------
        for pos in positionals:
            if isinstance(pos, Type):
                self._action.append(pos._type)
            elif isinstance(pos, Instance):
                self._action.append(to_var(pos))
                available_types.extend(pos._action.types)
                if last_pos_var:
                    self._graph._action(build.eq(last_pos_var, self._action.entity))
                last_pos_var = self._action.entity
            elif isinstance(pos, TypeUnion):
                self._action.append(to_var(pos()))
                available_types.extend([t._type for t in pos._types])
                if last_pos_var:
                    self._graph._action(build.eq(last_pos_var, self._action.entity))
                last_pos_var = self._action.entity
            elif isinstance(pos, Producer):
                self._action.append(to_var(pos))
                if last_pos_var:
                    self._graph._action(build.eq(last_pos_var, self._action.entity))
                last_pos_var = self._action.entity
            else:
                raise Exception(f"Unknown input type: {pos}")
        available_types.extend(self._action.types)
        if scope:
            available_types = [t for t in available_types if t.name.startswith(scope)]

        #--------------------------------------------------
        # Handle properties
        #--------------------------------------------------
        for name, val in named.items():
            prop = build.property_named(scope+name, available_types)

            if val is None:
                raise Exception(f"{prop}'s value is None, please provide a value for the property")

            prop_var = to_var(val)
            if is_collection(prop_var.value):
                raise Exception("Can't set a property to a collection")

            if not prop_var.name:
                prop_var.name = prop.name
            if action_type.is_effect():
                self._graph._check_property(prop)
            else:
                self._graph._check_property(prop, unknown_cardinality=True)
            self._action.append(prop, prop_var)

        #--------------------------------------------------
        # Entities
        #--------------------------------------------------
        self._var = self._action.entity
        if self._var.type == Builtins.Unknown and len(self._action.types):
            self._var.type = self._action.types[0]
        if not is_add and (self._action.types or self._action.bindings):
            self._add_to_graph()

    def _to_var(self):
        if not self._graph._stack.contains(self._context):
            exception = VariableOutOfContextException(Errors.call_source(), self._var.name)
            raise exception from None
        return self._var

    def _add_to_graph(self):
        for action in self._actions:
            self._graph._action(action)

    def __call__(self, *args, **kwargs):
        pass

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            return super().__setattr__(name, value)
        exception = InvalidPropertySetException(Errors.call_source(3))
        raise exception from None

    def _make_sub(self, name: str, existing=None):
        if self._sets.get(name) is not None:
            return self._sets[name]
        if existing is not None:
            return InstanceProperty(self._graph, self, name, var=existing._var, scope=self._scope)
        prop = build.property_named(self._scope+name, self._action.types)
        if self._action.bindings.get(prop):
            return InstanceProperty(self._graph, self, name, var=self._action.bindings[prop], scope=self._scope)
        return InstanceProperty(self._graph, self, name, scope=self._scope)

    def set(self, *args, **kwargs):
        if self._graph._stack.active() is self._context:
            self._sets.update(kwargs)
        Instance(self._graph, ActionType.Bind, [self, *args], kwargs, var=self._var, scope=self._scope)
        return self

    def persist(self, *args, **kwargs):
        Instance(self._graph, ActionType.Persist, [self, *args], kwargs, var=self._var, scope=self._scope)
        return self

    def unpersist(self, *args, **kwargs):
        Instance(self._graph, ActionType.Unpersist, [self, *args], kwargs, var=self._var, scope=self._scope)
        return self

#--------------------------------------------------
# InstanceProperty
#--------------------------------------------------

class InstanceProperty(Producer):
    def __init__(self, graph:'Graph', instance:Instance, name:str, var=None, scope:str=""):
        super().__init__(graph)
        self._instance = instance
        self._prop = build.property_named(scope+name, instance._action.types)
        self._var = var or Var(self._prop.type, name=name)
        self._check_context()
        self._scope = scope
        new = Instance(self._graph, ActionType.Get, [instance], {name: self._var}, scope=self._scope)
        self._action = new._action

    def _check_context(self):
        if not self._graph._stack.contains(self._instance._context):
            name = f"{self._instance._var.name}.{self._var.name}"
            exception = VariableOutOfContextException(Errors.call_source(), name, is_property=True)
            raise exception from None

    def __call__(self, *args, **kwargs):
        name = f"{self._instance._var.name}.{self._var.name}"
        exception = NonCallablePropertyException(Errors.call_source(), name)
        raise exception from None

    def _make_sub(self, name: str, existing=None):
        if existing is not None and self._graph._stack.is_active(existing._instance._context):
            return existing
        return getattr(Instance(self._graph, ActionType.Get, [self], {}), name)

    def _to_var(self):
        self._check_context()
        return self._var

    def or_(self, other):
        self._graph._remove_action(self._action)
        rel.pyrel_default(self._prop, other, self._instance, self)
        return self

    def in_(self, others):
        other_rel = InlineRelation(self._graph, [(x,) for x in others])
        return self == other_rel

    def has_value(self):
        self != rel.missing # type:ignore
        return self

    def _remove_if_unused(self):
        # When calling append/extend we aren't necessarily doing a get on the property,
        # but we will already have added one. If we're the only thing using this get,
        # we remove it so that it doesn't unnecessarily constrain the query.
        remove = True
        for item in reversed(self._graph._stack.items):
            if item is self._action:
                break
            elif isinstance(item, Action):
                if self._var in item.vars():
                    remove = False
                    break
        if remove:
            self._graph._remove_action(self._action)

    def set(self, *args, **kwargs):
        return Instance(self._graph, ActionType.Get, [self], {}).set(*args, **kwargs)

    def add(self, other):
        self._remove_if_unused()
        self._graph._check_property(self._prop, multi_valued=True)
        rel = Action(ActionType.Bind, to_var(self._instance), [], {self._prop: to_var(other)})
        self._graph._action(rel)

    def extend(self, others):
        self._remove_if_unused()
        self._graph._check_property(self._prop, True)
        for other in others:
            rel = Action(ActionType.Bind, to_var(self._instance), [], {self._prop: to_var(other)})
            self._graph._action(rel)

    def choose(self, num, unique=True):
        if num < 1:
            raise ValueError("Must choose a positive number of items")
        self._remove_if_unused()
        items = [getattr(Instance(self._graph, ActionType.Get, [self._instance], {}), self._prop.name) for ix in range(num)]
        if unique:
            for ix in range(num-1):
                items[ix] < items[ix+1]
        return items

#--------------------------------------------------
# Expression
#--------------------------------------------------

class Expression(Producer):
    def __init__(self, graph:'Graph', op:mType|Task, args:List[Any]):

        super().__init__(graph)
        self._var = None
        self._context = graph._stack.active()

        # For calls to tasks with known signatures, normalize their arguments by
        # throwing on missing inputs or constructing vars for missing outputs
        if op.properties and not op.isa(Builtins.Anonymous):
            for prop, arg in zip_longest(op.properties, args):
                if arg is None:
                    if prop.is_input:
                        raise TypeError(f"{op.name} is missing a required argument: '{prop.name}'")
                    else:
                        args.append(Var(prop.type, name=prop.name))

            # Expose the last output as the result, to ensure we don't double-create it in _use_var.
            # @NOTE: Literal values like 1 show up here from calls like `rel.range(0, len(df), 1)`
            if not op.properties[-1].is_input and isinstance(args[-1], Var):
                self._var = args[-1]

        self._expr = build.call(op, args)
        self._graph._action(self._expr)

    def __call__(self, *args, **kwargs):
        raise Exception("Expressions can't be called")

    def _use_var(self):
        if not self._var:
            self._var = Var(Builtins.Unknown)
            prop = build.property_named("result", self._expr.types)
            self._expr.append(prop, self._var)
        if not self._graph._stack.contains(self._context):
            exception = VariableOutOfContextException(Errors.call_source(), self._var.name or "a result")
            raise exception from None

    def _make_sub(self, name: str, existing=None):
        if existing is not None and existing._instance._context is self._graph._stack.active():
            return existing
        return getattr(Instance(self._graph, ActionType.Get, [self], {}), name)

#--------------------------------------------------
# RelationNS
#--------------------------------------------------

unsafe_symbol_pattern = re.compile(r"[^a-zA-Z0-9_]", re.UNICODE)
def safe_symbol(name: str):
    return f':"{name}"' if unsafe_symbol_pattern.search(name) else f":{name}"

class RelationNS():
    def __init__(self, ns:List[str], name:str, use_rel_namespaces=False):
        if name == "getdoc":
            rich.print("[red bold]GETDOC CALLED")
            traceback.print_stack()
            return
        self._name = name
        self._ns = ns
        self._subs = {}
        self._use_rel_namespaces = use_rel_namespaces
        self._rel = self._build_rel()

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        op = self._rel
        self._ensure_args(len(args))
        return Expression(get_graph(), op, list(args))

    def __getattr__(self, name: str) -> Any:
        self._subs[name] = self._make_sub(name, self._subs.get(name))
        return self._subs[name]

    def _make_sub(self, name: str, existing=None):
        if existing is not None:
            return existing
        ns = self._ns[:]
        if self._name:
            ns.append(self._name)
        return RelationNS(ns, name, use_rel_namespaces=self._use_rel_namespaces)

    def _build_rel(self, arg_count = 0):
        fqn_parts = self._ns + [self._name]
        if self._use_rel_namespaces:
            return build.relation('::'+'::'.join(fqn_parts), arg_count)
        if len(fqn_parts) == 1:
            return build.relation(fqn_parts[0], arg_count)
        else:
            return build.relation(f"{fqn_parts[0]}[{', '.join(safe_symbol(part) for part in fqn_parts[1:])}]", arg_count)

    def _ensure_args(self, arg_count):
        if len(self._rel.properties) <= arg_count:
            self._rel.properties = [Builtins.Relation.properties[i] for i in range(arg_count)]

    def add(self, *args):
        op = self._rel
        self._ensure_args(len(args))
        get_graph()._action(build.relation_action(ActionType.Bind, op, list(args)))

    def _to_var(self):
        return Var(Builtins.Relation, value=self._build_rel())

#--------------------------------------------------
# RawRelation
#--------------------------------------------------

class RawRelation(Producer):
    def __init__(self, graph:'Graph', name:str, arity:int):
        super().__init__(graph)
        self._name = name
        self._arity = arity
        self._type = build.relation(self._name, self._arity)

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return Expression(self._graph, self._type, list(args))

    def add(self, *args):
        self._graph._action(build.relation_action(ActionType.Bind, self._type, list(args)))

    def _make_sub(self, name: str, existing=None):
        return existing

    def _to_var(self):
        return Var(Builtins.Relation, value=self._type)

#--------------------------------------------------
# InlineRelation
#--------------------------------------------------

class InlineRelation():
    def __init__(self, graph:'Graph', data:List[Tuple]):
        self._var = Var(type=Builtins.InlineRawData)
        self._graph = graph
        cols = [[] for _ in range(len(data[0]))]
        for row in data:
            for i, val in enumerate(row):
                cols[i].append(to_var(val))

        params = [Var(value=col) for col in cols]
        params.append(self._var)
        q = build.relation_action(ActionType.Get, Builtins.InlineRawData, params)
        self._graph._action(q)

    def _to_var(self):
        return self._var

#--------------------------------------------------
# Symbol
#--------------------------------------------------

class Symbol():
    def __init__(self, name:str):
        self._var = Var(Builtins.Symbol, value=name)

    def _to_var(self):
        return self._var

#--------------------------------------------------
# RelationRef
#--------------------------------------------------

class RelationRef(Producer):
    def __init__(self, graph:'Graph', rel:Task|mType, args:List[Var]):
        super().__init__(graph)
        self._rel = rel
        self._args = args
        self._var = args[-1]
        self._action = build.relation_action(ActionType.Get, self._rel, self._args)

    def _use_var(self):
        self._graph._action(self._action)

    def _make_sub(self, name: str, existing=None):
        return getattr(Instance(self._graph, ActionType.Get, [self], {}), name)

    def __enter__(self):
        super().__enter__()
        self._use_var()

#--------------------------------------------------
# Export
#--------------------------------------------------

allowed_export_types = [Type, str, numbers.Number, datetime.datetime, datetime.date, bool]

def check_type(name, type):
    if not any(isinstance(type, t) or (inspect.isclass(type) and issubclass(type, t))
                for t in allowed_export_types):
        raise TypeError(f"Argument '{name}' is an unsupported type: {type}")

def export(model, schema, kwargs):
    def decorator(func):
        # Get type hints of the function
        hints = get_type_hints(func)
        input_types = [hints[name] for name in hints if name != 'return']
        arg_names = func.__code__.co_varnames[:func.__code__.co_argcount]
        for name in arg_names:
            if name not in hints:
                raise TypeError(f"Argument '{name}' must have a type hint")
            check_type(name, hints[name])

        output_types = []
        ret = hints.get('return')
        if typing.get_origin(ret) == tuple:
            for t in typing.get_args(ret):
                check_type("return", t)
                output_types.append(t)
        else:
            check_type("return", ret)
            output_types.append(ret)

        original_bytecode = Bytecode.from_code(func.__code__)
        new_bytecode = Bytecode()
        new_bytecode.argcount = func.__code__.co_argcount
        new_bytecode.argnames = func.__code__.co_varnames
        new_bytecode.docstring = func.__doc__
        new_bytecode.name = func.__name__

        for instr in original_bytecode:
            if isinstance(instr, Instr) and instr.name == "RETURN_VALUE":
                # Insert a call to the ret function before the return instruction
                if sys.version_info < (3, 11): # 3.10
                    new_bytecode.extend([
                        Instr("STORE_FAST", "______x"),
                        Instr("LOAD_GLOBAL", "ret"),
                        Instr("LOAD_FAST", "______x"),
                        Instr("CALL_FUNCTION", 1),
                    ])
                elif sys.version_info < (3, 12): # 3.11
                    new_bytecode.extend([
                        Instr("STORE_FAST", "______x"),
                        Instr("LOAD_GLOBAL", (True, "ret")),
                        Instr("LOAD_FAST", "______x"),
                        Instr("PRECALL", 0),
                        Instr("CALL", 1),
                    ])
                else: #3.12+
                    new_bytecode.extend([
                        Instr("STORE_FAST", "______x"),
                        Instr("LOAD_GLOBAL", (True, "ret")),
                        Instr("LOAD_FAST", "______x"),
                        Instr("CALL", 1),
                    ])
            else:
                new_bytecode.append(instr)

        # Create a new code object from the modified bytecode
        new_bytecode.append(Instr("RETURN_VALUE"))
        new_code = new_bytecode.to_code()

        # Create a new function from the new code object with the correct globals
        new_func = type(func)(new_code, func.__globals__, func.__name__, func.__defaults__, func.__closure__)

        # Update the globals dictionary of the new function to include 'ret'
        name = f"{schema}.{func.__name__}" if schema else func.__name__
        props = {"outputs": output_types, "name": name}
        ctx = Context(model, exec_type=TaskExecType.Procedure, props=props, format="snowpark", **kwargs)
        with ctx as ret:
            inputs = to_list(Vars(len(arg_names)))
            props["inputs"] = list(zip(arg_names, [to_var(i) for i in inputs], input_types))
            # Get the bytecode of the original function
            new_func.__globals__["ret"] = ret
            # Call the new function with the provided arguments
            new_func(*inputs)

        def wrapper():
            raise Exception("Exports can't be called directly. They are exported to the underlying platform")

        return wrapper
    return decorator

#--------------------------------------------------
# RuleStack
#--------------------------------------------------

class RuleStack():
    def __init__(self, graph:'Graph'):
        self.items = []
        self.stack = []
        self._graph = graph

    def push(self, item):
        self.stack.append(item)
        self.items.append(("push", item))

    def pop(self, item):
        self.stack.pop()
        self.items.append(("pop", item))
        if len(self.stack) == 0:
            compacted = self.compact()
            self.items.clear()
            if len(compacted.items):
                return compacted

    def contains(self, item):
        for i in self.stack:
            if i is item:
                return True

    def is_active(self, item):
        if isinstance(item, Context) and item._task.behavior != Behavior.Query:
            return False
        return item == self.active()

    def in_context(self, item):
        if isinstance(item, Context) and item._task.behavior != Behavior.Query:
            return False
        return self.contains(item)

    def active(self):
        try:
            cur = self.stack[-1]
            if cur is self._graph._temp_rule:
                exception = OutOfContextException(Errors.call_source())
                raise exception from None
            return cur
        except IndexError:
            exception = OutOfContextException(Errors.call_source())
            raise exception from None

    def _expression_start(self, buffer, single_use_vars):
        consume_from = -1
        if not len(buffer):
            return consume_from
        # we can only pull vars if their only use is for this condition
        used_vars = set(buffer[-1].requires_provides()[0] & single_use_vars)
        # walk buffer in reverse collecting vars in the action until we get one
        # that doesn't provide a var we care about
        for action in reversed(buffer[:-1]):
            if not isinstance(action, Action):
                break
            req, provs, _ = action.requires_provides()
            # don't pull in vars the represent root entities even though they're provided
            # by gets. This prevents scenarios where p = Person() would get pulled in if you
            # did with p.age > 10:
            provs = provs - {action.entity}
            if len(used_vars.intersection(provs)):
                used_vars.update(req & single_use_vars)
                consume_from -= 1
            else:
                break
        return consume_from

    def compact(self) -> Task:
        stack:List[Task] = []
        buffer = []

        var_uses = {}
        for item in self.items:
            if isinstance(item, Action):
                if item.action == ActionType.Get:
                    for var in item.vars():
                        var_uses[var] = var_uses.get(var, 0) + 1
                else:
                    for var in item.vars():
                        var_uses[var] = var_uses.get(var, 0) - 1

        # check for 2 refs - one create and one use
        single_use_vars = set([var for var, uses in var_uses.items() if uses >= 0])

        for item in self.items:
            if not isinstance(item, tuple):
                buffer.append(item)
                continue

            op, value = item
            if op == "push":
                if isinstance(value, Context):
                    if len(buffer):
                        stack[-1].items.extend(buffer)
                        buffer.clear()
                    task = value._task
                elif isinstance(value, RelationRef):
                    if len(buffer):
                        stack[-1].items.extend(buffer)
                        buffer.clear()
                    task = Task()

                elif isinstance(value, Producer):
                    consume_from = self._expression_start(buffer, single_use_vars)
                    stack[-1].items.extend(buffer[:consume_from])
                    buffer = buffer[consume_from:]
                    task = Task()
                else:
                    raise Exception(f"Unknown push type: {type(value)}")

                stack.append(task)

            elif op == "pop":
                cur = stack.pop()
                cur.items.extend(buffer)
                buffer.clear()
                if not len(stack):
                    return cur
                if isinstance(value, Context) and value._op:
                    stack[-1].items.append(build.call(value._op, [Var(value=value._args), Var(Builtins.Task, value=cur)]))
                else:
                    stack[-1].items.append(build.call(cur, list(cur.bindings.values())))

        raise Exception("No task found")

#--------------------------------------------------
# Graph
#--------------------------------------------------

locals = threading.local()
locals.graph_stack = []

def get_graph() -> 'Graph':
    _ensure_stack()
    if not len(locals.graph_stack):
        raise Exception("Outside of a model context")
    return locals.graph_stack[-1]

def _ensure_stack():
    if not hasattr(locals, "graph_stack"):
        locals.graph_stack = []
    return locals.graph_stack

rel = RelationNS([], "")
global_ns = RelationNS([], "", use_rel_namespaces=True)

def alias(ref:Any, name:str):
    var = to_var(ref)
    var.name = name
    return var

def Vars(count) -> Any:
    if count == 1:
        return Instance(get_graph(), ActionType.Get, [], {}, Var(Builtins.Unknown))
    return [Instance(get_graph(), ActionType.Get, [], {}, Var(Builtins.Unknown)) for _ in range(count)]

class Graph:
    def __init__(self, client:Client, name:str, format:str="default"):
        self.name = name
        self._executed = []
        self._client = client
        self._config = client.resources.config
        self.resources = client.resources
        self._prop_is_multi:Dict[str, bool] = {}
        self._format = format
        self._stack = RuleStack(self)
        self._temp_rule = Context(self)

        _ensure_stack().append(self)
        self._stack.push(self._temp_rule)

    #--------------------------------------------------
    # Rule stack
    #--------------------------------------------------

    def _flush_temp(self):
        if self._temp_rule:
            self._pop(self._temp_rule, is_temp=True)
            if not len(_ensure_stack()):
                _ensure_stack().append(self)
            self._temp_rule = None

    def _restore_temp(self):
        self._temp_rule = Context(self)
        _ensure_stack().append(self)
        self._stack.push(self._temp_rule)

    def _temp_is_active(self):
        return self._temp_rule and len(self._stack.items) > 1

    def _push(self, item):
        _ensure_stack().append(self)
        self._flush_temp()
        self._stack.push(item)

    def _pop(self, item, exec=True, is_temp=False):
        _ensure_stack().pop()
        task = self._stack.pop(item)
        try:
            if exec and task:
                self._exec(item, task)
        finally:
            if not is_temp and not len(self._stack.stack):
                self._restore_temp()

    def _action(self, action:Action|List[Action]):
        if isinstance(action, list):
            for a in action:
                self._action(a)
            return
        self._stack.items.append(action)

    def _remove_action(self, action):
        self._stack.items.remove(action)

    def _exec(self, context:Context, task):
        props = context._props
        if context._exec_type == TaskExecType.Rule:
            self._client.install(f"rule{len(self._executed)}", context._task)
        elif context._exec_type == TaskExecType.Query:
            context.results = self._client.query(context._task, tag=context._tag, format=context._format)
        elif context._exec_type == TaskExecType.Procedure:
            self._client.export_udf(props["name"], props["inputs"], props["outputs"], context._task, props.get("engine"))
        elif context._exec_type == TaskExecType.Export:
            self._client.export_table(props["name"], props["rel"], props["columns"], context._task, engine=props.get("engine"), refresh=props.get("refresh"))
        self._executed.append(context)

    #--------------------------------------------------
    # Property handling
    #--------------------------------------------------

    def _check_property(self, prop:mProperty, multi_valued=False, unknown_cardinality=False):
        name = prop.name
        if name in RESERVED_PROPS:
            exception = ReservedPropertyException(Errors.call_source(), name)
            raise exception from None

        if unknown_cardinality:
            return

        if name in self._prop_is_multi:
            if self._prop_is_multi[name] != multi_valued:
                raise Exception(f"Trying to use a property `{name}` as both singular and multi-valued")
            elif self._prop_is_multi[name] and Builtins.MultiValued not in prop.parents:
                prop.parents.append(Builtins.MultiValued)
        else:
            self._prop_is_multi[name] = multi_valued

        if not multi_valued and Builtins.FunctionAnnotation not in prop.parents:
            prop.parents.append(Builtins.FunctionAnnotation)
        elif multi_valued and Builtins.MultiValued not in prop.parents:
            prop.parents.append(Builtins.MultiValued)

    #--------------------------------------------------
    # Public API
    #--------------------------------------------------

    def Type(self, name:str, source=None):
        if source:
            return self.resources.to_model_type(self, name, source)
        return Type(self, name)

    def rule(self, **kwargs):
        return Context(self, **kwargs)

    def scope(self, **kwargs):
        return Context(self, **kwargs)

    def case(self, **kwargs):
        return Context(self, **kwargs)

    def match(self, multiple=False, **kwargs):
        if not multiple:
            return Context(self, behavior=Behavior.OrderedChoice, **kwargs)
        else:
            return Context(self, behavior=Behavior.Union, **kwargs)

    def query(self, **kwargs):
        return Context(self, exec_type=TaskExecType.Query, **kwargs)

    def export(self, object:str = "", **kwargs):
        return export(self, object, kwargs)

    def found(self, **kwargs):
        return Context(self, op=Builtins.Exists, **kwargs)

    def not_found(self, **kwargs):
        return Context(self, op=Builtins.Not, **kwargs)

    def union(self, **kwargs):
        return Context(self, behavior=Behavior.Union, **kwargs)

    def ordered_choice(self, **kwargs):
        return Context(self, behavior=Behavior.OrderedChoice, **kwargs)

    def read(self, name:str, **kwargs):
        from relationalai.loaders.loader import read_resource_context # We do the late import to break an dependency cycle
        return read_resource_context(self, name, **kwargs)

    def load_raw(self, path:str):
        if os.path.isfile(path):
            if path.endswith('.rel'):
                self._client.load_raw_file(path)
        elif os.path.isdir(path):
            for root, _, files in os.walk(path):
                for file in files:
                    if file.endswith('.rel'):
                        file_path = os.path.join(root, file)
                        self._client.load_raw_file(file_path)

    def exec_raw(self, code:str, readonly=True, raw_results=True, inputs={}):
        return self._client.exec_raw(code, readonly=readonly, raw_results=raw_results, inputs=inputs)

    def install_raw(self, code:str, name:str|None=None):
        self._client.install_raw(code, name)



from typing import List, Set
from contextlib import contextmanager
import pprint
import keyword
import re
import unicodedata
from relationalai import metamodel as mm
from relationalai.clients.test import Document, Install, Query
from gentest.gen import ir

pp = pprint.PrettyPrinter(indent=2, depth=2)

def sanitize_variable_name(input_string):
    input_string = unicodedata.normalize('NFKD', input_string)
    # Replace non-alphanumeric characters with underscores
    sanitized = re.sub(r'\W|^(?=\d)', '_', input_string)

    # Check if the resulting string is a keyword and append an underscore if it is
    if keyword.iskeyword(sanitized):
        sanitized += '_'

    return sanitized


class Scope:
    def __init__(self, parent: 'Scope|None' = None):
        self.mappings: dict[mm.Type|mm.Var, str] = {}
        self.parent = parent
        self.unused_properties: dict[mm.Var, bool] = {}
        self.introduced: set[mm.Var] = set()
        self.used_entity_names: set[str] = set()

    def child(self) -> 'Scope':
        return Scope(self)

    def get(self, of: mm.Type|mm.Var) -> str|None:
        existing = self.mappings.get(of)
        if existing:
            return existing
        if self.parent:
            return self.parent.get(of)

    def constant(self, value: mm.Value) -> str:
        if isinstance(value, list):
            str_repr = f"[{', '.join(self.value(item) for item in value)}]"
        else:
            str_repr = repr(value)
            if str_repr == "nan" or str_repr == "inf" or str_repr == "-inf":
                if isinstance(value, float):
                    str_repr = f"float(\"{str_repr}\")"

        return str_repr

    # @FIXME: When do I ever _not_ want the name unique-ifying of as_root?
    def var(self, of: mm.Type|mm.Var, as_root = True, uses = True) -> str:
        if of in self.unused_properties and uses:
            del self.unused_properties[of]
        existing = self.get(of)
        if existing:
            return existing

        match of:
            case mm.Type():
                str_repr = sanitize_variable_name(of.name)
            case mm.Var():
                if of.value is not None:
                    str_repr = self.constant(of.value)
                elif as_root:
                    base = of.type.name.lower()
                    ix = 1
                    cur = base
                    while cur in self.used_entity_names:
                        ix += 1
                        cur = f"{base}{ix}"

                    self.used_entity_names.add(cur)
                    str_repr = sanitize_variable_name(cur)
                else:
                    str_repr = sanitize_variable_name(f"{of.type.name.lower()}_{str(of)}" )
            case _:
                assert of is None

        self.mappings[of] = str_repr
        return str_repr

    def value(self, of:mm.Var|mm.Value, as_entity = False):
        match of:
            case mm.Var():
                return self.var(of, as_entity)
            case _:
                return self.constant(of)

    def map_property(self, subject: mm.Var, property: mm.Property, var: mm.Var):
        assert not self.mappings.get(var)
        mapped = f"{self.var(subject)}.{sanitize_variable_name(property.name)}"
        self.mappings[var] = mapped

class Builder:
    def __init__(self, document: 'DocumentEmitter', scope: Scope, joiner = "\n"):
        self.doc = document
        self.scope = scope
        self.chunks: List[str] = []
        self.joiner = joiner

    def chunk(self, text: str):
        self.chunks.append(text)

    def stringify(self, joiner: str|None = None):
        joiner = (joiner if joiner else self.joiner)
        return joiner.join(self.chunks)

    @contextmanager
    def nest(self):
        nested = Builder(self.doc, self.scope.child(), f"{self.joiner}    ")
        try:
            yield nested
        finally:
            self.chunk(f"    {nested.stringify()}")

    @contextmanager
    def call(self, op: str, store_to: str|None = None):
        nested = Builder(self.doc, self.scope, joiner = ", ")
        try:
            yield nested
        finally:
            expr = f"{op}({nested.stringify()})"
            if store_to:
                self.chunk(f"{store_to} = {expr}")
            else:
                self.chunk(expr)



class DocumentEmitter:
    @classmethod
    def from_document(cls, doc: Document):
        new = cls()
        for block in doc.blocks:
            match block:
                case Install():
                    new.rule(block.task)
                case Query():
                    new.query(block.task)
                case _:
                    raise Exception(f"Unhandled block type {block}")

        return new

    def __init__(self, graph_ctor: str = "rai.clients.test.Graph(\"test\")", is_nested = False):
        self.graph_name = "graph"
        self.graph_ctor = graph_ctor

        self.scope = Scope()
        self.blocks: List[Builder] = []
        self.uses_types: Set[mm.Type] = set()

        self.is_nested = is_nested
        self.sep = "\n    " if is_nested else "\n"

    def rule(self, task: mm.Task):
        builder = Builder(self, self.scope, self.sep)
        builder.chunk(f"with {self.graph_name}.rule():")
        emit_task(task, builder)
        if len(builder.chunks) > 1:
            self.blocks.append(builder)
            return builder

    def query(self, task: mm.Task):
        builder = Builder(self, self.scope, self.sep)
        builder.chunk(f"with {self.graph_name}.query() as select:")
        emit_task(task, builder)
        if len(builder.chunks) > 1:
            self.blocks.append(builder)
            return builder

    def use(self, type: mm.Type):
        self.uses_types.add(type)

    def stringify(self):

        body = self.sep + self.sep.join([
            "# Types",
            *[f"{self.scope.var(type)} = {self.graph_name}.Type(\"{type.name}\")" for type in self.uses_types],
            "",

            "# Rules",
            *[f"{block.stringify()}\n" for block in self.blocks]
        ])

        return "\n".join([
            "import relationalai as rai",
            "",

            "# Init",
            f"{self.graph_name} = {self.graph_ctor}" if not self.is_nested else f"with {self.graph_ctor} as {self.graph_name}:",
            body,
        ])




def emit_task(task: mm.Task, into: Builder):
    with into.nest() as body:
        for item in task.items:
            emit_action(item, body)

        for v in body.scope.unused_properties:
            body.chunk(body.scope.var(v, uses=False))

def emit_action(action: mm.Action, into: Builder):
    match action.action:
        case mm.ActionType.Bind | mm.ActionType.Persist | mm.ActionType.Unpersist | mm.ActionType.Construct:
            if action.entity.value == mm.Builtins.Return:
                emit_return_action(action, into)
            else:
                emit_assert_action(action, into)
        case mm.ActionType.Get:
            emit_get_action(action, into)
        case mm.ActionType.Call:
            emit_call_action(action, into)

        case _:
            raise Exception(f"Unhandled action type in emitter '{action.action}'.")

def emit_return_action(action: mm.Action, into: Builder):
    # @FIXME: Need to know the actual name of the select here and choose what to name the return var
    with into.call("select", "response") as call:
        for type in action.types:
            call.doc.use(type)
            call.chunk(call.scope.var(type))
        for var in action.bindings.values():
            call.chunk(call.scope.var(var))

def emit_assert_action(action: mm.Action, into: Builder):
    # @FIXME: Check if var is A) subject of a make_identity call, and B) not already provided by another assert/get
    #         to switch to op = "add"
    creates_entity = action.entity not in into.scope.introduced

    match action.action:
        case mm.ActionType.Bind:
            op = "add" if creates_entity else "set"
        case mm.ActionType.Persist:
            op = "persist"
        case  mm.ActionType.Unpersist:
            op = "unpersist"
        case _:
            raise Exception(f"Invalid assert action: {action.action}")

    types = action.types

    if creates_entity:
        subject_type = types[0]
        assert subject_type

        into.doc.use(subject_type)
        op = f"{into.scope.var(subject_type)}.{op}"
        types = types[1:]
    else:
        op = f"{into.scope.var(action.entity, True)}.{op}"

    with into.call(op, into.scope.var(action.entity, True) if creates_entity and not action.entity.value else None) as call:
        for type in types:
            call.doc.use(type)
            call.chunk(call.scope.var(type))

        for (prop, var) in action.bindings.items():
            call.chunk(f"{sanitize_variable_name(prop.name)}={call.scope.var(var)}")


def emit_get_action(action: mm.Action, into: Builder):
    creates_entity = action.entity not in into.scope.introduced

    subject_type = action.types[0]
    assert subject_type

    into.doc.use(subject_type)

    store_to = None
    if creates_entity:
        into.scope.introduced.add(action.entity)
        store_to = into.scope.var(action.entity, True)

    with into.call(into.scope.var(subject_type), store_to) as call:
        if not creates_entity:
            call.chunk(call.scope.var(action.entity, True))

        for type in action.types[1:]:
            into.doc.use(type)
            call.chunk(call.scope.var(type))

        for (prop, var) in action.bindings.items():
            call.scope.map_property(action.entity, prop, var)
            call.scope.unused_properties[var] = True


OP_NAME_MAP = {
    "=": "==",
    "average": "avg",

}

def emit_call_action(action: mm.Action, into: Builder):
    task = action.entity.value
    assert isinstance(task, mm.Task)

    is_infix = ir.implements(task, mm.Builtins.Infix)  # task.name in rel.rel_infix

    if is_infix:
        op = OP_NAME_MAP.get(task.name, task.name)
        p1 = task.properties[0]
        p2 = task.properties[1]
        expr = f"{into.scope.var(action.bindings[p1])} {op} {into.scope.var(action.bindings[p2])}"
        if len(task.properties) > 2:
            res_prop = task.properties[2]
            res = action.bindings[res_prop]
            expr = f"{into.scope.var(res)} = {expr}"

        into.chunk(expr)
    elif task.name == "make_identity":
        pass # this is implicitly handled by the scope.introduced set right now.
        # print(f"@TODO: Handle make_identity in call {str(action)}")
    elif ir.implements(task, mm.Builtins.Aggregate):
        op = OP_NAME_MAP.get(task.name, task.name)
        op = f"{into.doc.graph_name}.aggregates.{op}"
        last_prop = task.properties[-1]
        store_to = action.bindings.get(last_prop, None) if not last_prop.is_input else None # @FIXME: More robust result management
        with into.call(op, into.scope.var(store_to, as_root=True) if store_to else None) as call:
            # @FIXME: agg property names don't match the compiler expected kwargs
            # This all goes  away if group is renamed to per and projection to args (or vice versa in the public agg fns)
            projection = next(prop for prop in task.properties if prop.name == "projection")
            proj_arg = action.bindings.get(projection, None)
            if proj_arg and proj_arg.value and isinstance(proj_arg.value, list):
                for v in proj_arg.value:
                    call.chunk(call.scope.value(v))

            group = next(prop for prop in task.properties if prop.name == "group")
            group_arg = action.bindings.get(group, None)
            if group_arg and group_arg.value:
                call.chunk(f"per={call.scope.var(group_arg)}")
    else:
        op = OP_NAME_MAP.get(task.name, task.name)
        last_prop = task.properties[-1]
        store_to = action.bindings.get(last_prop, None) if not last_prop.is_input else None # @FIXME: More robust result management
        with into.call(op, into.scope.var(store_to) if store_to else None) as call:
            for prop in task.properties:
                if prop == last_prop and store_to:
                    continue
                arg = call.scope.var(action.bindings[prop])
                call.chunk(f"{prop.name}={arg}")



#pyright: reportUnusedExpression=false
import relationalai as rai
from relationalai.clients.snowflake import Snowflake

# This test expects the model to already be present, with the data loaded into it.
# It's the same model used by the TastyBytes benchmark.
model = rai.Model(name="PyRelTastyBytesBenchmark", config=globals().get("config"))
sf = Snowflake(model)

Record = sf.TASTYBYTES.HARMONIZED.LOYALTY_ORDERS_REGION_CALIFORNIA

# TODO: re-enable when performance improves (i.e. by turning off IVM, etc)
# with model.query() as select:
#     record = Record()
#     num_records = aggregates.count(record)
#     select(num_records)



import datetime as dt

import pytest
import relationalai as rai
from relationalai import std

alias = std.alias

model = rai.Model(name=globals().get("name", "test_dates"), config=globals().get("config"))

# Check that 'date' works.
# DataFrame should have one row for each case.
with model.query() as select:
    with model.match(multiple=True) as matched:
        with model.case():
            date = std.dates.date(2024, 1, 1)
            std.dates.Date(date)
            matched.add(date, id=1)
        with model.case():
            ord = dt.date(2024, 1, 1).toordinal()
            date = std.dates.date.fromordinal(ord)
            std.dates.Date(date)
            matched.add(date, id=2)
        with model.case():
            _dt = dt.datetime(2024, 1, 1, tzinfo=dt.timezone.utc)
            date = std.dates.date.fromdatetime(_dt)
            std.dates.Date(date)
            matched.add(date, id=3)
        with model.case():
            _dt = std.dates.datetime(2024, 1, 1)
            date = std.dates.date.fromdatetime(_dt)
            std.dates.Date(date)
            matched.add(date, id=4)
    select(matched.id, alias(matched, "date"))

# Check that 'datetime' works.
# DataFrame should have one row for each scope in the union.
with model.query() as select:
    with model.match(multiple=True) as matched:
        with model.case():
            datetime = std.dates.datetime(2024, 1, 1, 6, 30, 0, 0, "UTC")
            std.dates.DateTime(datetime)
            matched.add(datetime, id=1)
        with model.case():
            ord = dt.date(2024, 1, 1).toordinal()
            datetime = std.dates.datetime.fromordinal(ord)
            std.dates.DateTime(datetime)
            matched.add(datetime, id=2)
        with model.case():
            _date = dt.date(2024, 1, 1)
            datetime = std.dates.datetime.fromdate(_date)
            std.dates.DateTime(datetime)
            matched.add(datetime, id=3)
        with model.case():
            _date = std.dates.date(2024, 1, 1)
            datetime = std.dates.datetime.fromdate(_date)
            std.dates.DateTime(datetime)
            matched.add(datetime, id=4)
    select(matched.id, alias(matched, "datetime"))

# Check that 'date_add' and 'date_subtract' work.
# DataFrame should have one row for each case.
for func in [
    std.dates.date_add,
    std.dates.date_subtract
]:
    with model.query() as select:
        with model.match(multiple=True) as matched:
            with model.case():
                date = dt.date(2024, 1, 1)
                p = std.dates.days(1)
                matched.add(func(date, period=p), id=1)
            with model.case():
                date = dt.datetime(2024, 1, 1, 6, 30, 45, tzinfo=dt.timezone.utc)
                p = std.dates.weeks(1)
                matched.add(func(date, period=p), id=2)
            with model.case():
                date = std.dates.date(2024, 1, 1)
                p = std.dates.months(1)
                matched.add(func(date, period=p), id=3)
            with model.case():
                date = std.dates.datetime(2024, 1, 1, 6, 30, 45)
                p = std.dates.years(1)
                matched.add(func(date, period=p), id=4)
        select(matched.id, alias(matched, "date"))

# Check that date parts work.
for func in [
    std.dates.year,
    std.dates.month,
    std.dates.day,
    std.dates.week,
    std.dates.weekday,
    std.dates.isoweekday,
    std.dates.quarter,
    std.dates.dayofyear,
]:
    with model.query() as select:
        with model.match(multiple=True) as matched:
            with model.case():
                date = dt.date(2024, 1, 1)
                matched.add(func(date), id=1)
            with model.case():
                date = std.dates.date(2024, 1, 1)
                matched.add(func(date), id=2)
        select(matched.id, alias(matched, func.__name__))

    # Check that passing an invalid argument raises a TypeError.
    with pytest.raises(TypeError):
        with model.query():
            func("INVALID")

# Check that datetime parts work.
for func in [
    std.dates.year,
    std.dates.month,
    std.dates.day,
    std.dates.hour,
    std.dates.minute,
    std.dates.second,
    std.dates.week,
    std.dates.weekday,
    std.dates.isoweekday,
    std.dates.quarter,
    std.dates.dayofyear,
]:
    with model.query() as select:
        with model.match(multiple=True) as matched:
            with model.case():
                datetime = dt.datetime(2024, 1, 1, 6, 30, 45, tzinfo=dt.timezone.utc)
                matched.add(func(datetime), id=1)
            with model.case():
                datetime = std.dates.datetime(2024, 1, 1, 6, 30, 45)
                matched.add(func(datetime), id=2)
        select(matched.id, alias(matched, func.__name__))

    # Check that passing an invalid argument raises a TypeError.
    with pytest.raises(TypeError):
        with model.query():
            func("INVALID")

# Check that period constructors work
with model.query(dynamic=True) as select:
    time = std.dates.datetime(2024, 1, 1)
    with model.match(multiple=True, dynamic=True) as matched:
            for (ix, period) in enumerate([
                std.dates.microseconds,
                std.dates.milliseconds,
                std.dates.seconds,
                std.dates.minutes,
                std.dates.hours,
                std.dates.days,
                std.dates.weeks,
                std.dates.months,
                std.dates.years,
            ]):
                with model.case():
                    matched.add(std.dates.date_add(time, period(1)), id=ix)
    select(matched.id, matched)

# Check that the date_range function works.
for kwargs in [
    {"start": dt.date(2024, 1, 1), "end": dt.date(2024, 1, 7)},
    {"start": dt.datetime(2024, 1, 1, tzinfo=dt.timezone.utc), "end": dt.datetime(2024, 1, 2, tzinfo=dt.timezone.utc), "freq": "H"},
    {"start": dt.date(2024, 1, 1), "end": dt.date(2024, 1, 2), "freq": "H"},
    {"start": dt.datetime(2024, 1, 1, tzinfo=dt.timezone.utc), "periods": 12, "freq": "M"},
    {"end": dt.date(2024, 1, 7), "periods": 7, "freq": "D"},
]:
    with model.query() as select:
        dates = std.dates.date_range(**kwargs)
        select(dates)



import io
import logging
import os
from pathlib import Path
from typing import Callable

import rich
from rich.syntax import Syntax

from relationalai import debugging
from relationalai.debugging import logger
from relationalai.errors import RelQueryError

def path_to_slug(path: Path, base_path:str|Path):
    return str(path.relative_to(base_path)).replace("/", "__").replace(".py", "")

class QueryTextLogger(logging.Handler):
    def __init__(self):
        super().__init__()
        # compilation events that get 'results' added to them
        self.blocks = []

    def emit(self, record):
        d = record.msg
        if isinstance(d, dict):
            if d["event"] == "compilation":
                self.blocks.append({**d})
            elif d["event"] == "time" and d["type"] == "query" and not d.get("internal"):
                self.blocks[-1]["results"] = d["results"]

def validate_block_snapshots(
        file_path: Path,
        snapshot,
        get_snapshot_str: Callable, # given a block, return the snapshot string
        snapshot_prefix:str,
        model_kwargs:dict|None = None
    ):
    with open(file_path, "r") as file:
        code = file.read()
        # @TODO: Consider suppressing stdout

        # install logger to capture Rel compilation
        query_logger = QueryTextLogger()
        logger.addHandler(query_logger)

        # ensure that the exec knows what file is being run
        model_kwargs = model_kwargs or {}
        model_kwargs["__file__"] = str(file_path)
        code_object = compile(code, file_path, 'exec')
        exec(code_object, model_kwargs, None)

        logger.removeHandler(query_logger)

        block_index = 0
        for block in query_logger.blocks:
            try:
                cmp_string = get_snapshot_str(block)
                if cmp_string is not None:
                    snapshot.assert_match(cmp_string, f"{snapshot_prefix}{block_index}.txt")
                    # only counting blocks that actually assert something
                    block_index += 1
            except RelQueryError as err:
                err.pprint()
                raise err from None
            except AssertionError as err:
                lines = str(err).splitlines()
                if len(lines) < 3:
                    raise
                header, info, _, *body = lines
                with io.StringIO() as buf:
                    console = rich.console.Console(file=buf, force_terminal=True)
                    console.print(header)
                    console.print(info)

                    source_info = debugging.get_source(block["task"])
                    assert source_info and source_info.line is not None
                    source = debugging.find_block_in(code, source_info.line, str(file_path))

                    console.print()
                    base = os.getcwd()
                    console.print("In", f"./{file_path.relative_to(base)}" if file_path.is_relative_to(base) else file_path)
                    if source.source:
                        console.print(Syntax(source.source, "python", dedent=True, line_numbers=True, start_line=source.line, padding=1))

                    console.print('\n'.join(body))
                    raise Exception(buf.getvalue()) from None
